import copy

from deepproblog.dataset import (
    DataLoader,
    QueryDataset,
    MutatingDataset,
    NumericIncorrectOutputMutator,
    NoiseMutatorDecorator,
)
from deepproblog.query import Query
from deepproblog.utils import parse
from problog.logic import Term, Constant


def test_query_dataset(tmpdir):
    with tmpdir.join("queries.txt").open(mode="wt") as tempfile:
        tempfile.write(
            "a(2,3,5).\nsubs(a(3,3,b),[[b,123]]).\n0.5 :: subs(a(3,3,c),[[c,123]]).\n0.7 :: b(1,2,3)."
        )
    dataset_file = QueryDataset(tmpdir.join("queries.txt"))
    dataset_list = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
            Query(parse("a(3,3,c)."), substitution={Term("c"): Constant(123)}, p=0.5),
            Query(parse("b(1,2,3)."), p=0.7),
        ]
    )
    assert len(dataset_file) == 4
    assert len(dataset_list) == 4
    assert dataset_file.queries == dataset_list.queries
    assert dataset_list.to_queries() == dataset_list.queries


def test_dataset_write_to_file(tmpdir):
    dataset_list = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
        ]
    )
    with tmpdir.join("queries_out.txt").open(mode="wt") as tempfile:
        dataset_list.write_to_file(tempfile)
    # Test that we can reload it.
    dataset_reloaded = QueryDataset(tmpdir.join("queries_out.txt"))
    assert dataset_reloaded.queries == dataset_list.queries


def test_subset():
    dataset = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
        ]
    )
    assert dataset.subset(1).to_queries() == [dataset.queries[0]]


def test_extension():
    dataset1 = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
        ]
    )
    dataset2 = QueryDataset([Query(parse("a(1,2,3)."))])
    assert (dataset1 + dataset2).to_queries() == dataset1.queries + dataset2.queries


def test_mutating_dataset():
    dataset1 = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
        ]
    )

    def mutator(_: int, q: Query):
        q2 = copy.copy(q)
        q2.visited = True
        return q2

    mutated = MutatingDataset(dataset1, mutator)
    assert all(hasattr(e, "visited") for e in mutated.to_queries())


def test_noise_mutator():
    """Test that the noise is approximately correct"""
    hit_count = 0

    def inner_mutator(_: int, q: Query):
        nonlocal hit_count
        hit_count += 1
        return q

    mutator = NoiseMutatorDecorator(p=0.75, inner_mutator=inner_mutator, seed=123)

    total_count = 2000
    for i in range(total_count):
        mutator(i, Query(Term("dummy")))

    assert 0.7 < hit_count / total_count < 0.8

    # Check we get the same result twice
    hit_count1 = hit_count
    hit_count = 0
    for i in range(total_count):
        mutator(i, Query(Term("dummy")))
    assert hit_count == hit_count1


def test_numeric_incorrect_output_mutator():
    mutator = NumericIncorrectOutputMutator(domain=list(range(10)), seed=123)
    r1 = mutator(1, Query(parse("a(1,2,3).")))
    r2 = mutator(1, Query(parse("a(1,2,3).")))
    r3 = mutator(2, Query(parse("a(1,2,3).")))
    r4 = mutator(2, Query(parse("a(1,2,3).")))
    assert r1 == r2
    assert r3 == r4
    assert r1 != r3
    assert r1.query.args[-1].value != 3
    assert r3.query.args[-1].value != 3


def test_dataloader():
    dataset = QueryDataset(
        [
            Query(parse("a(2,3,5).")),
            Query(parse("a(1,2,3).")),
            Query(parse("a(3,3,b)."), substitution={Term("b"): Constant(123)}),
        ]
    )
    loader = DataLoader(dataset, 2, False)
    one_epoch = list(loader)
    assert one_epoch[0] + one_epoch[1] == dataset.queries



from deepproblog.utils import check_path

template = """
[Default]
batch_size = {0}
infoloss = {1}
name = poker_batch_{0}_infoloss_{1}
"""
i = 0

check_path("parameter_cfg/0.cfg")

for batch_size in [10, 25, 50, 100]:
    for infoloss in [0, 0.5, 1.0, 2.0, 4.0]:
        with open("parameter_cfg/{}.cfg".format(i), "w") as f:
            f.write(template.format(batch_size, infoloss))
        i += 1

template = """
[Default]
batch_size = 10
infoloss = 0.5
labeled = {0}
name = poker_batch_labeled_{0}_{1}
"""
i = 0

check_path("experiment/0.cfg")

for labeled in [300]:
    for rep in range(10):
        with open("experiment/{}.cfg".format(i), "w") as f:
            f.write(template.format(labeled, rep))
        i += 1



import torch

from deepproblog.dataset import DataLoader, QueryDataset
from deepproblog.engines import ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.Forth import EncodeModule
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model

train = 2
test = 8

train_queries = QueryDataset("data/train{}_test{}_train.txt".format(train, test))
test_queries = QueryDataset("data/train{}_test{}_test.txt".format(train, test))
val = QueryDataset("data/train{}_test{}_dev.txt".format(train, test))


net1 = EncodeModule(30, 50, 10, "tanh")
network1 = Network(net1, "neural1")
network1.optimizer = torch.optim.Adam(net1.parameters(), lr=0.02)

net2 = EncodeModule(22, 10, 2, "tanh")
network2 = Network(net2, "neural2")
network2.optimizer = torch.optim.Adam(net2.parameters(), lr=0.02)

model = Model("choose.pl", [network1, network2])
test_model = Model(
    "choose.pl", [Network(net1, "neural1", k=1), Network(net2, "neural2", k=1)]
)
model.set_engine(ExactEngine(model), cache=True)
test_model.set_engine(ExactEngine(test_model), cache=False)
train_obj = train_model(
    model,
    DataLoader(train_queries, 50),
    40,
    log_iter=20,
    test=lambda x: [("Accuracy", get_confusion_matrix(test_model, val).accuracy())],
    test_iter=100,
)



import torch.nn as nn


class MNIST_CNN(nn.Module):
    def __init__(self):
        super(MNIST_CNN, self).__init__()
        self.encoder = nn.Sequential(
            nn.Conv2d(1, 6, 5),
            nn.MaxPool2d(2, 2),  # 6 24 24 -> 6 12 12
            nn.ReLU(True),
            nn.Conv2d(6, 16, 5),  # 6 12 12 -> 16 8 8
            nn.MaxPool2d(2, 2),  # 16 8 8 -> 16 4 4
            nn.ReLU(True),
        )

    def forward(self, x):
        x = x.unsqueeze(0)
        x = self.encoder(x)
        x = x.view(-1, 16 * 4 * 4)
        return x


class MNIST_Classifier(nn.Module):
    def __init__(self, n=10, with_softmax=True):
        super(MNIST_Classifier, self).__init__()
        self.with_softmax = with_softmax
        if with_softmax:
            self.softmax = nn.Softmax(1)
        self.classifier = nn.Sequential(
            nn.Linear(16 * 4 * 4, 120),
            nn.ReLU(),
            nn.Linear(120, 84),
            nn.ReLU(),
            nn.Linear(84, n),
        )

    def forward(self, x):
        x = self.classifier(x)
        if self.with_softmax:
            x = self.softmax(x)
        return x.squeeze(0)


class MNIST_Net(nn.Module):
    def __init__(self, n=10, with_softmax=True, size=16 * 4 * 4):
        super(MNIST_Net, self).__init__()
        self.with_softmax = with_softmax
        self.size = size
        if with_softmax:
            if n == 1:
                self.softmax = nn.Sigmoid()
            else:
                self.softmax = nn.Softmax(1)
        self.encoder = nn.Sequential(
            nn.Conv2d(1, 6, 5),
            nn.MaxPool2d(2, 2),  # 6 24 24 -> 6 12 12
            nn.ReLU(True),
            nn.Conv2d(6, 16, 5),  # 6 12 12 -> 16 8 8
            nn.MaxPool2d(2, 2),  # 16 8 8 -> 16 4 4
            nn.ReLU(True),
        )
        self.classifier = nn.Sequential(
            nn.Linear(size, 120),
            nn.ReLU(),
            nn.Linear(120, 84),
            nn.ReLU(),
            nn.Linear(84, n),
        )

    def forward(self, x):
        x = self.encoder(x)
        x = x.view(-1, self.size)
        x = self.classifier(x)
        if self.with_softmax:
            x = self.softmax(x)
        return x



from random import randint

import torch
from problog.logic import Constant

from deepproblog.dataset import DataLoader
from deepproblog.dataset import NoiseMutatorDecorator, MutatingDataset
from deepproblog.engines import ExactEngine
from deepproblog.examples.MNIST.data import MNISTOperator, MNIST_train, MNIST_test
from deepproblog.examples.MNIST.network import MNIST_Net
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.optimizer import SGD
from deepproblog.query import Query
from deepproblog.train import train_model


def noise(_, query: Query):
    new_query = query.replace_output([Constant(randint(0, 18))])
    return new_query


dataset = MNISTOperator(
    dataset_name="train",
    function_name="addition_noisy",
    operator=sum,
    size=1,
)
noisy_dataset = MutatingDataset(dataset, NoiseMutatorDecorator(0.2, noise))
queries = DataLoader(noisy_dataset, 2)

network = MNIST_Net()
net = Network(network, "mnist_net", batching=True)
net.optimizer = torch.optim.Adam(network.parameters(), lr=1e-3)
model = Model("models/noisy_addition.pl", [net])

model.add_tensor_source("train", MNIST_train)
model.add_tensor_source("test", MNIST_test)

model.set_engine(ExactEngine(model))
model.optimizer = SGD(model, 1e-3)

train = train_model(model, queries, 1, log_iter=100)



import torch
import torch.nn as nn

from deepproblog.utils.standard_networks import MLP


class EncodeModule(nn.Module):
    def __init__(self, in_size, mid_size, out_size, activation=None):
        super().__init__()
        if activation == "tanh":
            self.mlp = MLP(in_size, mid_size, out_size, activation=nn.Tanh)
        else:
            self.mlp = MLP(in_size, mid_size, out_size)
        self.in_size = in_size

    def __call__(self, *x):
        input = torch.zeros(self.in_size)
        for j, i in enumerate(x):
            i = int(i)
            input[j * 10 + i] = 1
        return self.mlp(input)



from typing import Mapping, Iterator

import torch
import torchvision
import torchvision.transforms as transforms
from problog.logic import Term, Constant

from deepproblog.dataset import Dataset
from deepproblog.query import Query

transform = transforms.Compose(
    [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
)

datasets = {
    "train": torchvision.datasets.MNIST(
        root='data/', train=True, download=True, transform=transform
    ),
    "test": torchvision.datasets.MNIST(
        root='data/', train=False, download=True, transform=transform
    ),
}


class MNISTImages(Mapping[Term, torch.Tensor]):

    def __iter__(self) -> Iterator:
        for i in range(self.dataset):
            yield self.dataset[i][0]

    def __len__(self) -> int:
        return len(self.dataset)

    def __init__(self, subset):
        self.subset = subset
        self.dataset = datasets[self.subset]

    def __getitem__(self, item):
        return self.dataset[int(item[0])][0]


class AdditionDataset(Dataset):

    def __init__(self, subset):
        self.subset = subset
        self.dataset = datasets[subset]

    def __len__(self):
        return len(self.dataset) // 2

    def to_query(self, i: int) -> Query:
        image1 = Term("tensor", Term(self.subset, Constant(i * 2)))
        image2 = Term("tensor", Term(self.subset, Constant(i * 2 + 1)))
        label = Constant(int(self.dataset[i*2][1] + self.dataset[i*2+1][1]))
        term = Term('addition', image1, image2, label)
        return Query(term)



import pytest
from deepproblog.engines import ExactEngine

from deepproblog.model import Model
from deepproblog.query import Query
from problog.logic import Term, Var

def _create_model(program) -> Model:
    """Setup code: Load a program minimally"""
    model = Model(program, [], load=False)
    engine = ExactEngine(model)
    model.set_engine(engine)
    return model



def test_tensor_index():
    program = """
P :: a(I) :-  list_to_tensor([0.25,0.1,0.4,0.15], Tensor), between(0,3,I), tensor_index(Tensor,[I],P).
    """
    model = _create_model(program)
    q = Query(Term("a", Var("X")))
    r = model.solve([q])[0].result
    expected_result = [0.25, 0.1, 0.4, 0.15]
    for k in r:
        i = int(k.args[0])
        assert pytest.approx(expected_result[i]) == float(r[k])


def test_less_than():
    program = """
    a :- list_to_tensor([0.8],T1), list_to_tensor([0.2],T2), less_than(T1,T2).
    b :- list_to_tensor([0.2],T1), list_to_tensor([0.8],T2), less_than(T1,T2).
    """
    model = _create_model(program)
    q1 = Query(Term("a"))
    q2 = Query(Term("b"))
    assert pytest.approx(0.0) == model.solve([q1])[0].result[Term("a")]
    assert pytest.approx(1.0) == model.solve([q2])[0].result[Term("b")]


from unittest import TestCase

import pytest
prolog_engine = pytest.importorskip("deepproblog.engines.prolog_engin")

from prolog_engine import PrologEngine
from prolog_engine.heuristics import PartialProbability
from problog import get_evaluatable
from problog.logic import Term, Constant, Var, Clause
from problog.program import PrologString

ucs = PartialProbability()


class TestSWIProgram(TestCase):
    def compare_query_probability(self, engine, program, query, target):
        ground = engine.ground(program, query, label="query")
        sdd = get_evaluatable().create_from(ground)
        p = sdd.evaluate()[query]
        self.assertAlmostEqual(target, p)

    def has_solution(self, engine, program, query, answer):
        ground = engine.ground(program, query, label="query")
        sdd = get_evaluatable().create_from(ground)
        p = sdd.evaluate()
        print(p)
        self.assertAlmostEqual(p[answer], 1.0)

    def test_add_clause(self):
        program = """
a :- b, c.
0.5::a.
0.5::b.
0.5::c.
0.8::d.
query(a).
"""
        program = PrologString(program)
        engine = PrologEngine(k=10, heuristic=ucs, exploration=False)
        program = engine.prepare(program)

        self.compare_query_probability(engine, program, Term("a"), 0.625)
        clause_i = program.add_clause(Clause(Term("a"), Term("d")))[1]
        self.compare_query_probability(engine, program, Term("a"), 0.925)
        program.remove_clause(clause_i)
        self.compare_query_probability(engine, program, Term("a"), 0.625)

    def test_non_ground_is(self):
        program = "b(X,Y) :- Y is X + 1."
        program = PrologString(program)
        engine = PrologEngine(k=5, heuristic=ucs, exploration=False)
        program = engine.prepare(program)

        self.has_solution(
            engine,
            program,
            Term("b", Constant(2), Var("Y")),
            Term("b", Constant(2), Constant(3)),
        )



import csv
from ast import literal_eval
from collections import defaultdict
from pathlib import Path
from random import choices
from re import match

import torch
from torch.utils.data import Dataset as PyDataset

from typing import Union
from deepproblog.dataset import Dataset
from deepproblog.query import Query
from problog.logic import Term, Constant, list2term

# 1 noise_free
# 2 supporting
# 3 irrelevant
# 4 disconnected

dataset_names = {
    "noise_0": "train 1.2, 1.3, test 1.2,1.3,2.3,3.3,4.3",
    "noise_1": "train 2.2, 2.3, test 1.3,2.2,2.3,3.3,4.3",
    "noise_2": "train 3.2, 3.3, test 1.3,2.3,3.2,3.3,4.3",
    "noise_3": "train 4.2, 4.3, test 1.3,2.3,3.3,4.2,4.3",
    "sys_gen_0": "train 1.2 test 1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,1.10",
    "sys_gen_1": "train 1.2,1.3 test 1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,1.10",
    "sys_gen_2": "train 1.2,1.3,1.4 test 1.2,1.3,1.4,1.5,1.6,1.7,1.8,1.9,1.10",
}

synonyms = {
    "sys_gen_0": "data/data_a9dcffad",
    "sys_gen_1": "data/data_b11e6a4f/",
    "sys_gen_2": "data/data_f70b574f",
    "noise_0": "data/data_2d5007e7/",
    "noise_1": "data/data_a7d9402e/",
    "noise_2": "data/data_6b1c2f15/",
    "noise_3": "data/data_47b0ffea/",
}

labels1 = [
    "daughter",
    "daughter_in_law",
    "mother",
    "mother_in_law",
    "sister",
    "sister_in_law",
    "grandmother",
    "granddaughter",
    "niece",
    "aunt",
    "wife",
]
labels2 = [
    "son",
    "son_in_law",
    "father",
    "father_in_law",
    "brother",
    "brother_in_law",
    "grandfather",
    "grandson",
    "nephew",
    "uncle",
    "husband",
]
labels3 = [
    "child",
    "child_in_law",
    "parent",
    "parent_in_law",
    "sibling",
    "sibling_in_law",
    "grandparent",
    "grandchild",
    "nephew",
    "uncle",
    "so",
]

labels = labels1 + labels2


def remove_gender(x):
    if x in labels3:
        return x
    try:
        return labels3[labels1.index(x)]
    except ValueError:
        return labels3[labels2.index(x)]


class ListDataset(PyDataset):
    def __init__(self, data):
        PyDataset.__init__(self)
        self.data = data

    def __getitem__(self, item):
        return self.data[item]

    def __len__(self):
        return len(self.data)


def parse_relation(x):
    x = x.replace("-", "_")
    if x == "neice":
        x = "niece"
    return x


class Story(object):
    def __init__(self, line=None, vocab=None, copy=None, edges=True):
        self.use_edges = edges
        self.positive = True
        if copy is not None:
            self.edges = copy.edges
            self.query = copy.query
            self.answer = copy.answer
            self.text = copy.text
            self.entities = copy.entities
        else:
            text, query, answer = line[2], line[3], line[5]
            story_edges, edge_types, genders = line[11], line[12], line[14]
            story_edges = literal_eval(story_edges)
            edge_types = literal_eval(edge_types)
            edge_types = [parse_relation(x) for x in edge_types]
            try:
                self.edges = {
                    story_edges[m]: edge_types[m] for m in range(len(story_edges))
                }
            except IndexError:
                self.edges = {}

            self.query = tuple(line.strip("'").lower() for line in query[1:-1].split(", "))
            answer = parse_relation(answer)
            if answer not in labels:
                raise Exception(answer + " not in labels")
            self.answer = answer

            genders = [g.split(":")[0].lower() for g in genders.strip().split(",")]

            for punctuation in [".", ",", "'", "!", ";"]:
                text = text.replace(punctuation, " " + punctuation + " ")
            self.text = [w for w in text.strip().split(" ") if len(w) > 0]

            self.entities = {name: ind for ind, name in enumerate(genders)}
            for i in range(len(self.text)):
                if self.text[i][0] == "[":
                    word = self.text[i][1:-1].lower()
                    j = self.entities[word]
                    self.text[i] = j
                else:
                    self.text[i] = self.text[i].lower()
                    word = self.text[i].lower()
                    vocab[word] += 1

    def get_sentences(self):
        sentences = [[]]
        entities = [set()]
        for word in self.text:
            if word in [".", "!"]:
                sentences.append([])
                entities.append(set())
            else:
                if type(word) is int:
                    entities[-1].add(word)
                    sentences[-1].append(str(word))
                else:
                    sentences[-1].append(word)

        return [
            (list(entities[i]), sentences[i])
            for i in range(len(sentences))
            if len(sentences[i]) > 0
        ]

    def get_negatives(self, nr_negatives, gender):
        if nr_negatives < 1:
            return []
        possible_corruptions = [
            x for x in (labels if gender else labels3) if x != self.answer
        ]
        negative_labels = choices(possible_corruptions, k=nr_negatives)
        negatives = []
        for label in negative_labels:
            copy = Story(copy=self)
            copy.answer = label
            copy.positive = False
            negatives.append(copy)
        return negatives

    def remove_gender(self):
        self.answer = remove_gender(self.answer)
        self.edges = {k: remove_gender(self.edges[k]) for k in self.edges}

    def to_query(self, query_type):
        if query_type == "split":
            return self.to_query_split()
        elif query_type == "edges":
            return self.to_query_edges()
        else:
            return self.to_query_whole()

    def get_text(self):
        return " ".join(str(x) for x in self.text)

    def get_relations(self):
        return self.edges

    def to_query_edges(self):
        entities = [Constant(x) for x in sorted(self.entities.values())]
        x, y = (
            Constant(self.entities[self.query[0]]),
            Constant(self.entities[self.query[1]]),
        )

        edges = [
            Term(self.edges[e], Constant(e[0]), Constant(e[1])) for e in self.edges
        ]

        query_term = Term(
            "clutrr_edges",
            list2term(entities),
            list2term(edges),
            x,
            Term(self.answer),
            y,
        )
        return Query(query_term, output_ind=[-2], p=float(self.positive))

    def to_query_whole(self):
        entities = [Constant(x) for x in sorted(self.entities.values())]
        substitution = dict()
        x, y = (
            Constant(self.entities[self.query[0]]),
            Constant(self.entities[self.query[1]]),
        )
        substitution[Term("text")] = Constant(
            '"' + " ".join(str(w) for w in self.text) + '"'
        )
        query_term = Term(
            "query_rel",
            Term("s", list2term(entities), Term("text")),
            x,
            Term(self.answer),
            y,
        )
        return Query(query_term, substitution, output_ind=[-2], p=float(self.positive))

    def to_query_split(self):
        entities = [Constant(x) for x in sorted(self.entities.values())]
        x, y = (
            Constant(self.entities[self.query[0]]),
            Constant(self.entities[self.query[1]]),
        )
        sentences = []
        for e, s in self.get_sentences():
            e = list2term([Constant(x) for x in e])
            s = Constant(" ".join(s))
            sentences.append(Term("s", e, s))
        query_term = Term(
            "clutrr_text",
            list2term(entities),
            list2term(sentences),
            x,
            Term(self.answer),
            y,
        )
        return Query(query_term, output_ind=[-2], p=float(self.positive))


class CLUTRR_Dataset(Dataset):
    def __init__(self, name, stories, vocab, gender, query_type, nr_negatives=0):
        self.gender = gender
        self.name = name
        self.type = query_type
        self.vocab = vocab
        negatives = []
        for story in stories:
            negatives += story.get_negatives(nr_negatives, gender)
        self.stories = stories + negatives

    def __getitem__(self, item):
        if not self.gender:
            self.stories[item].remove_gender()
        return self.stories[item]

    def to_query(self, i, j=None):
        if not self.gender:
            self.stories[i].remove_gender()
        return self.stories[i].to_query(self.type)

    def __len__(self):
        return len(self.stories)

    def __repr__(self):
        return dataset_names[self.name]


class CLUTRR(object):
    def __init__(self, name):
        self.synonym = name
        name = synonyms.get(name, name)
        self.name = Path(name)
        self.vocab = defaultdict(int)
        self.data = dict()
        self.subsets = self.name.glob("*.csv")
        for subset in self.subsets:
            self.load_file(subset)
        self.vocab = (
                ["ENT{}".format(x) for x in range(10)]
                + ["ENT", "OOV"]
                + sorted(self.vocab, key=lambda x: self.vocab[x], reverse=True)
        )

    def load_file(self, filename):
        with open(filename) as f:
            reader = csv.reader(f)
            next(reader)
            self.data[filename.stem] = [Story(line, self.vocab) for line in reader]

    def get_vocabulary(self):
        return {w: i for i, w in enumerate(self.vocab)}

    def save_vocabulary(self, path):
        with open(path, "w") as f:
            f.write("\n".join(self.vocab))

    def load_vocabulary(self, path):
        with open(path) as f:
            self.vocab = [word.rstrip("\n") for word in f]

    def get_embeddings(self, path="glove.6B.50d.txt", n=50):
        weights = torch.empty((len(self.vocab), n))
        weights.normal_(0, 1)
        vocab = self.get_vocabulary()
        with open(path) as f:
            for line in f:
                line = line.strip().split(" ")
                word, vec = line[0], [float(x) for x in line[1:]]
                try:
                    i = vocab[word]
                    weights[i, :] = torch.tensor(vec)
                except KeyError:
                    pass
        return weights

    def get_dataset(self, pattern, separate=False, **kwargs) -> Union[dict[CLUTRR_Dataset], CLUTRR_Dataset]:
        if separate:
            datasets = dict()
            for k in self.data:
                if match(pattern, k):
                    datasets[k] = CLUTRR_Dataset(
                        self.synonym, self.data[k], self.get_vocabulary(), **kwargs
                    )
            return datasets
        else:
            stories = []
            for k in self.data:
                if match(pattern, k):
                    stories += self.data[k]
            return CLUTRR_Dataset(
                self.synonym, stories, self.get_vocabulary(), **kwargs
            )

    def get_relations(self, pattern, gender):
        data = []
        for k in self.data:
            if match(pattern, k):
                for story in self.data[k]:
                    if gender is False:
                        story.remove_gender()
                    entities = list(story.entities.values())
                    data.append((story.get_text(), entities, story.get_relations()))
        return data



import json
import os
import pickle
import time
from collections import defaultdict
from io import BytesIO
from os import PathLike
from random import random
from typing import (
    Collection,
    Optional,
    Tuple,
    List,
    Mapping,
    Any,
    Sequence,
    Union,
    IO,
    Callable,
)
from zipfile import ZipFile

import torch
from problog.logic import term2list, Term, Clause, Constant, InstantiationError, Or
from problog.program import PrologString, PrologFile, SimpleProgram, LogicProgram

from deepproblog.engines.builtins import register_tensor_predicates
from .dataset import Dataset, ImageDataset
from .embeddings import TermEmbedder
from .engines import Engine
from .network import Network
from .optimizer import Optimizer
from .query import Query
from .semiring import Result
from .solver import Solver
from .utils import check_path


class Model(object):
    def __init__(
        self,
        program_string: Union[str, os.PathLike],
        networks: Collection[Network],
        embeddings: Optional[TermEmbedder] = None,
        load: bool = True,
    ):
        """

        :param program_string: A string representing a DeepProbLog program or the path to a file containing a program.
        :param networks: A collection of networks that will be used to evaluate the neural predicates.
        :param embeddings: A TermEmbedder used to embed Terms in the program.
        :param load: If true, then it will attempt to load the program from 'program_string',
         else, it will consider program_string to be the program itself.
        """
        self.networks = dict()
        if load:
            self.program: LogicProgram = PrologFile(str(program_string))
        else:
            self.program: LogicProgram = PrologString(program_string)
        self.parameters = []
        self.parameter_groups = []
        self._extract_parameters()
        for network in networks:
            self.networks[network.name] = network
            network.model = self
        self.solver: Optional[Solver] = None
        self.eval_mode = False
        self.embeddings = embeddings
        self.tensor_sources = dict()
        self.optimizer = Optimizer(self)

    def get_embedding(self, term: Term):
        return self.embeddings.get_embedding(term)

    def evaluate_nn(self, to_evaluate: List[Tuple[Term, Term]]):
        """
        :param to_evaluate: List of neural predicates to evaluate
        :return: A dictionary with the elements of to_evaluate as keys, and the output of the NN as values.
        """
        result = dict()
        evaluations = defaultdict(list)
        # Group inputs per net to send in batch
        for net_name, inputs in to_evaluate:
            net = self.networks[str(net_name)]
            if net.det:
                tensor_name = Term("nn", net_name, inputs)
                if tensor_name not in self.solver.engine.tensor_store:
                    evaluations[net_name].append(inputs)
            else:
                if inputs in net.cache:
                    result[(net_name, inputs)] = net.cache[inputs]
                    del net.cache[inputs]
                else:
                    evaluations[net_name].append(inputs)
        for net in evaluations:
            network = self.networks[str(net)]
            out = network([term2list(x, False) for x in evaluations[net]])
            for i, k in enumerate(evaluations[net]):
                if network.det:
                    tensor_name = Term("nn", net, k)
                    self.solver.engine.tensor_store.store(out[i], tensor_name)
                else:
                    result[(net, k)] = out[i]
        return result

    def set_engine(self, engine: Engine, **kwargs):
        """
        Initializes the solver of this model with the given engine and additional arguments.
        :param engine: The engine that will be used to ground queries in this model.
        :param kwargs: Additional arguments passed to the solver.
        :return:
        """
        self.solver = Solver(self, engine, **kwargs)
        register_tensor_predicates(engine)

    def solve(self, batch: Sequence[Query]) -> List[Result]:
        return self.solver.solve(batch)

    def ground_dataset(self, dataset: Dataset):
        total_time = 0
        compile_times = []
        ground_times = []
        for q in dataset.to_queries():
            start = time.time()
            result = self.solver.cache.get(q)
            total_time += time.time() - start
            if not result.from_cache:
                compile_times.append(result.compile_time)
                ground_times.append(result.ground_time)
        return {
            "total_time": total_time,
            "ground_times": ground_times,
            "compile_times": compile_times,
        }

    def save_state(self, filename: Union[str, PathLike, IO[bytes]], complete=False):
        """
        Saves the state of this model to a zip file with the given filename. This only includes the probabilistic
            parameters and all parameters of the neural networks, but not the model architecture or neural architectures
        :param filename: The filename to save the model to.
        :param complete: If true, save neural networks with information needed to resume training.
        :return:
        """
        check_path(filename)
        with ZipFile(filename, "w") as zipf:
            with zipf.open("parameters", "w") as f:
                pickle.dump(self.parameters, f)
            for n in self.networks:
                with zipf.open(n, "w") as f:
                    self.networks[n].save(f, complete=complete)

    def load_state(self, filename: Union[str, PathLike, IO[bytes]]):
        """
        Restore the state of this model from the given filename. This only includes the probabilistic parameters
            and all parameters of the neural networks, but not the model architecture or neural architectures.
        :param filename: The filename to restore the model from.
        :return:
        """
        with ZipFile(filename) as zipf:
            with zipf.open("parameters") as f:
                self.parameters = pickle.load(f)
            for n in self.networks:
                with zipf.open(n) as f:
                    self.networks[n].load(BytesIO(f.read()))

    def eval(self):
        """
        Set the mode of all networks in the model to eval.
        """
        self.eval_mode = True
        for n in self.networks:
            self.networks[n].eval()
        self.solver.engine.eval()

    def train(self):
        """
        Set the mode of all networks in the model to train.
        :return:
        """
        self.eval_mode = False
        for n in self.networks:
            self.networks[n].train()
        self.solver.engine.train()

    def register_foreign(
        self, func: Callable, function_name: str, arity_in: int, arity_out: int
    ):
        self.solver.engine.register_foreign(func, function_name, arity_in, arity_out)

    def __str__(self):
        return "\n".join(str(line) for line in self.program)

    def get_tensor(self, term: Term) -> torch.Tensor:
        """

        :param term: A term of the form tensor(_).
        If the tensor is of the form tensor(a(*args)), then it well look into tensor source a.
        :return:  Returns the stored tensor identifier by the term.
        """
        if len(term.args) > 0 and term.args[0].functor in self.tensor_sources:
            return self.tensor_sources[term.args[0].functor][term.args[0].args]
        return self.solver.get_tensor(term)

    def store_tensor(self, tensor: torch.Tensor) -> Term:
        """
        Stores a tensor in the tensor store and returns and identifier.
        :param tensor: The tensor to store.
        :return: The Term that is the identifier by which this tensor can be uniquely identified in the logic.
        """
        return Term("tensor", Constant(self.solver.engine.tensor_store.store(tensor)))

    def add_tensor_source(
        self, name: str, source: Union[ImageDataset, Mapping[Any, torch.Tensor]]
    ):
        """
        Adds a named tensor source to the model.
        :param name: The name of the added tensor source.
        :param source: The tensor source to add
        :return:
        """
        self.tensor_sources[name] = source

    def get_hyperparameters(self) -> dict:
        """
        Recursively build a dictionary containing the most important hyperparameters in the model.
        :return: A dictionary that contains the values of the most important hyperparameters of the model.
        """
        parameters = dict()
        parameters["solver"] = (
            None if self.solver is None else self.solver.get_hyperparameters()
        )
        parameters["networks"] = [
            self.networks[network].get_hyperparameters() for network in self.networks
        ]
        parameters["program"] = self.program.to_prolog()
        return parameters

    def hyperparameters_to_file(self, filename):
        """
        Write the output of the get_hyperparameter() method in JSON format to a file.
        :param filename: The path to write the hyperparameters to.
        :return:
        """
        with open(filename, "w") as f:
            f.write(json.dumps(self.get_hyperparameters()))

    def _extract_parameters(self):
        translated = SimpleProgram()
        for n in self.program:
            if type(n) is Term:
                if (
                    n.probability is not None
                    and type(n.probability) is Term
                    and n.probability.functor == "t"
                ):
                    i = self._add_parameter(n.probability.args[0])
                    p = n.probability.with_args(Constant(i))
                    n = n.with_probability(p)
                translated.add_statement(n)
            elif type(n) is Clause:
                if (
                    n.head.probability is not None
                    and type(n.head.probability) is Term
                    and n.head.probability.functor == "t"
                ):
                    i = self._add_parameter(n.head.probability.args[0])
                    p = n.head.probability.with_args(Constant(i))
                    head = n.head.with_probability(p)
                    n = Clause(head, n.body)
                translated.add_statement(n)
            elif type(n) is Or:
                new_list = []
                new_group = []
                for x in n.to_list():
                    if (
                        x.probability is not None
                        and type(x.probability) is Term
                        and x.probability.functor == "t"
                    ):
                        i = self._add_parameter(x.probability.args[0])
                        new_group.append(i)
                        p = x.probability.with_args(Constant(i))
                        new_list.append(x.with_probability(p))
                    else:
                        new_list.append(x)
                if len(new_group) > 0:
                    self.parameter_groups.append(new_group)
                n = Or.from_list(new_list)
                translated.add_statement(n)
            else:
                translated.add_statement(n)
        self.program = translated

    def _add_parameter(self, val: Constant):
        i = len(self.parameters)
        try:
            val = float(val)
        except InstantiationError:
            val = random()
        self.parameters.append(val)
        return i



from pathlib import Path
from typing import Optional

from problog.engine import GenericEngine
from problog.formula import LogicFormula
from problog.logic import Term, Var, Constant
from problog.program import LogicProgram
from pyswip import Prolog

from .heuristics import Heuristic
from .swi_program import SWIProgram

root = Path(__file__).parent


def escape_strings_in_term(term: Term):
    if type(term) is Term:
        return Term(term.functor, *(escape_strings_in_term(x) for x in term.args))
    elif type(term) is Var:
        return term
    elif type(term) is Constant:
        if term.is_string():
            value = term.value
            if value[0] != '"' and value[-1] != '"':
                return Constant('"' + value + '"')
        return term
    raise ValueError("{} not handled in escape_strings_in_term".format(type(term)))


class PrologEvaluationException(Exception):
    """Exception from PrologEngine for unexpected result when evaluating a query."""


class PrologEngine(GenericEngine):
    def __init__(
        self,
        k,
        heuristic: Optional[Heuristic],
        exploration: bool,
        timeout=None,
        ignore_timeout=False,
    ):
        super().__init__()
        self.k = k
        self.heuristic = heuristic
        self.prolog = Prolog()
        self.timeout = timeout
        self.ignore_timeout = ignore_timeout
        self.exploration = exploration
        path = root / "prolog_files" / "engine_heap.pl"
        self.prolog.consult(path.as_posix())

    def prepare(self, db):
        program = SWIProgram(db, heuristic=self.heuristic)
        return program

    def ground(self, sp: LogicProgram, term, target=None, label=None, *args, **kwargs):
        if type(sp) != SWIProgram:
            sp = self.prepare(sp)
        if target is None:
            target = LogicFormula(keep_all=True)
        proofs = self.get_proofs(term, sp)
        result = sp.add_proof_trees(proofs, target=target, label=label)
        return result

    def ground_all(
        self,
        sp: LogicProgram,
        target=None,
        queries=None,
        evidence=None,
        *args,
        **kwargs,
    ):
        if type(sp) != SWIProgram:
            sp = self.prepare(sp)
        if target is None:
            target = LogicFormula()
        if queries is None:
            queries = [
                q[0].args[0]
                for q in self.ground(
                    sp, Term("query", Var("X")), *args, **kwargs
                ).queries()
            ]
        for q in queries:
            self.ground(sp, q, target, *args, **kwargs)
        return target

    def get_proofs(self, q: Term, program: SWIProgram, profile=0):
        exploration = "true" if self.exploration else "false"
        q_term = escape_strings_in_term(q)
        query_str = "prove({},{},Proofs,{},{})".format(
            q_term, self.k, self.heuristic.name, exploration
        )
        if self.timeout is not None:
            query_str = "call_with_time_limit({},{})".format(self.timeout, query_str)
        try:
            res = program.query(query_str, profile=profile)
        except TimeoutError:
            if self.ignore_timeout:
                return []
            else:
                raise TimeoutError()
        except OverflowError:
            return []
        if len(res) != 1:
            raise PrologEvaluationException(
                f"Expected exactly 1 result, got {len(res)}"
            )
        return res[0]["Proofs"]



import math

import torch
from torch import nn as nn


class JSD(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, eps=1e-8):
        logN = math.log(float(x.shape[0]))

        y = torch.mean(x, 0)
        y = y * (y + eps).log() / logN
        y = y.sum()

        x = x * (x + eps).log() / logN
        x = x.sum(1).mean()
        return 1.0 - x + y


class InfoLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, eps=1e-8):
        x = torch.mean(x, 0)
        logN = math.log(float(x.shape[0]))
        x = x * (x + eps).log() / logN
        neg_entropy = x.sum()
        return 1.0 + neg_entropy


class EntropyLoss(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x, eps=1e-8):
        logN = math.log(float(x.shape[0]))
        x = x * (x + eps).log() / logN
        neg_entropy = x.sum(1)
        return -neg_entropy.mean()


infoLoss = InfoLoss()
entropyLoss = EntropyLoss()
jsd = JSD()


def MSE(p, target):
    return (p - target) ** 2, 2 * (p - target)



import torch

from deepproblog.dataset import DataLoader, QueryDataset
from deepproblog.engines import ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.Forth import EncodeModule
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model

train = 2
test = 8

train_queries = QueryDataset("data/train{}_test{}_train.txt".format(train, test))
dev_queries = QueryDataset("data/train{}_test{}_dev.txt".format(train, test))
test_queries = QueryDataset("data/train{}_test{}_test.txt".format(train, test))

fc1 = EncodeModule(20, 20, 2)

model = Model(
    "compare.pl",
    [Network(fc1, "swap_net", optimizer=torch.optim.Adam(fc1.parameters(), 1.0))],
)
model.set_engine(ExactEngine(model), cache=True)

test_model = Model("compare.pl", [Network(fc1, "swap_net", k=1)])
test_model.set_engine(ExactEngine(test_model), cache=False)

train_obj = train_model(
    model,
    DataLoader(train_queries, 16),
    40,
    log_iter=50,
    test_iter=len(train_queries),
    test=lambda x: [
        ("Accuracy", get_confusion_matrix(test_model, dev_queries).accuracy())
    ],
)



from time import time
from typing import Optional, List, Tuple, Dict, Type, TYPE_CHECKING

from problog.formula import LogicDAG, LogicFormula
from problog.logic import Term, term2list
from problog.sdd_formula import SDD

from deepproblog.semiring import Result
from deepproblog.semiring import Semiring

if TYPE_CHECKING:
    from deepproblog.model import Model


class ArithmeticCircuit(object):
    def __init__(
        self,
        formula: LogicFormula,
        semiring: Optional[Type[Semiring]],
        ground_time: Optional[float] = None,
        sdd_auto_gc: bool = False,
    ):
        """
        :param formula: The ground logic formula that will be compiled.
        :param semiring: Factory method that creates the semiring
        :param ground_time: Optional. Keep track of time it took to ground out formula. Used for timing statistics.
        :param sdd_auto_gc: Controls if the PySDD auto-GC feature should be turned on (may be needed for large problems)
        """
        start = time()
        self.proof = LogicDAG.create_from(formula, keep_named=True)
        self.sdd = SDD.create_from(self.proof, sdd_auto_gc=sdd_auto_gc)
        self.compile_time = time() - start
        self.re_evaluate = False
        self.semiring = semiring
        self.ground_time = ground_time
        self.results = dict()

    def evaluate(
        self,
        model: "Model",
        substitution: Optional[dict] = None,
        re_evaluate: bool = True,
    ) -> Result:
        """
        Evaluates the arithmetic circuit.
        :param model: The model (neural network and parameters) to evaluate the AC in.
        :param substitution: Optional dict. The substitution applied to the parameterized AC. See apply_term.
        :param re_evaluate: Force re-evaluation of the neural networks
        :return:
        """
        substitution = dict() if substitution is None else substitution
        values = None
        if self.re_evaluate or re_evaluate:
            to_evaluate = self.extract_neural()
            to_evaluate = [(e[0], e[1].apply_term(substitution)) for e in to_evaluate]
            values = model.evaluate_nn(to_evaluate)
        self.re_evaluate = True
        semiring = self.semiring(model, substitution, values)
        start = time()
        evaluation = self.sdd.evaluate(semiring=semiring)
        if substitution is not None:
            evaluation = {k.apply_term(substitution): evaluation[k] for k in evaluation}
        eval_time = time() - start
        return Result(
            evaluation,
            semiring,
            self.ground_time,
            self.compile_time,
            eval_time,
            self.proof,
        )

    def extract_neural(self) -> List[Tuple[Term, Term]]:
        """
        :return: Returns a set of all neural predicates that need to be evaluated.
        """
        neural_eval = []
        weights = self.sdd.get_weights()
        for w in weights:
            w = weights[w]
            if type(w) is Term:
                if w.functor == "nn":
                    self._add_ordered_evaluation(w.args[0], w.args[1], neural_eval)

        return neural_eval

    def _add_ordered_evaluation(self, name, arguments, evals):
        # Check arguments for tensors that need to be evaluated
        for argument in term2list(arguments, deep=False):
            if argument.functor == "tensor":
                argument = argument.args[0]
                if argument.functor == "nn":
                    self._add_ordered_evaluation(*argument.args, evals)
        k = (name, arguments)
        if k not in evals:
            evals.append(k)

    def get_named(self) -> Dict[Term, int]:
        """
        :return: A dictionary mapping all named nodes in the SDD to their node id.
        """
        named = dict()
        for n in self.sdd.get_names():
            named[n[0]] = n[1]
        return named

    def save(self, filename):
        manager = self.sdd.get_manager().get_manager()
        i = list(self.sdd.queries())[0][1] - 1
        inode = self.sdd.get_manager().nodes[i]
        constraint_inode = self.sdd.get_constraint_inode()
        node = self.sdd.get_manager().conjoin(inode, constraint_inode)
        manager.minimize()
        var_names = [
            (var, self.sdd.get_node(atom).probability)
            for var, atom in self.sdd.var2atom.items()
        ]

        manager.vtree().save((filename + ".vtree").encode())
        manager.save((filename + ".sdd").encode(), node)
        with open(filename + ".tsv", "w") as f:
            f.write("\n".join(str(v) + "\t" + str(p) for v, p in var_names))



import torch
import torch.nn as nn
import torch.optim as optim

from deepproblog.utils import count_parameters
from deepproblog.utils.standard_networks import MLP

vocab = dict()
hidden_size = 512
embed_size = 256

with open("data/vocab_746.txt") as f:
    for i, word in enumerate(f):
        word = word.strip()
        # vocab[i] = word
        vocab[word] = i


def tokenize(sentence):
    sentence = sentence.split(" ")
    tokens = []
    numbers = list()
    indices = list()
    for i, word in enumerate(sentence):
        if word.isdigit():
            numbers.append(int(word))
            tokens.append("<NR>")
            indices.append(i)
        else:
            if word in vocab:
                tokens.append(word)
            else:
                tokens.append("<UNK>")
    return [vocab[token] for token in tokens], numbers, indices


class RNN(nn.Module):
    def __init__(self, vocab_size, hidden_size, p_drop=0.0):
        super(RNN, self).__init__()
        self.lstm = nn.GRU(
            embed_size, hidden_size, 1, bidirectional=True, batch_first=True
        )
        self.hidden_size = hidden_size
        self.embedding = nn.Embedding(
            vocab_size, embed_size
        )  # , _weight=torch.nn.Parameter(weights))
        self.dropout = nn.Dropout(p_drop)

    def forward(self, sentence):
        sentence = sentence.value.strip('"')
        x, _, indices = tokenize(sentence)
        n1, n2, n3 = indices
        seq_len = len(x)
        x = torch.LongTensor(x).unsqueeze(0)
        x = self.embedding(x)
        x, _ = self.lstm(x)
        x = x.view(seq_len, 2, self.hidden_size)
        x1 = torch.cat([x[-1, 0, ...], x[n1, 0, ...], x[n2, 0, ...], x[n3, 0, ...]])
        x2 = torch.cat([x[0, 1, ...], x[n1, 1, ...], x[n2, 1, ...], x[n3, 1, ...]])
        x = torch.cat([x1, x2])
        #        return x
        return self.dropout(x)


def get_networks(lr, p_drop, n=8):
    rnn = RNN(len(vocab), hidden_size, p_drop=p_drop)
    network1 = MLP(hidden_size * n, 6)
    network2 = MLP(hidden_size * n, 4)
    network3 = MLP(hidden_size * n, 2)
    network4 = MLP(hidden_size * n, 4)

    print("rnn params", count_parameters(rnn))
    print("net1 params", count_parameters(network1))

    names = ["rnn", "nn_permute", "nn_op1", "nn_swap", "nn_op2"]
    networks = [rnn, network1, network2, network3, network4]

    networks = [
        (networks[i], names[i], optim.Adam(networks[i].parameters(), lr=lr))
        for i in range(5)
    ]

    return networks



from abc import ABC, abstractmethod
from typing import List, Optional, Collection

import torch
from problog.logic import Term


class TermEmbedder(ABC):
    """
    Generic embedder for terms.
    """

    @abstractmethod
    def get_embedding(self, term: Term) -> torch.Tensor:
        """

        :param term: Term to be embedded
        :return:
        """
        pass

    def zero_grad(self):
        """
        Zeroes the gradients for the learnable tensors.
        """
        pass

    def step(self):
        """
        Perform calls the step method for the optimizers of the contained learned embeddings.
        """
        pass


class OneHotEncoding(TermEmbedder):
    """
    One-hot embedder. Assign each unique term a unique one-hot embedding.
    """

    def __init__(self, n: int):
        """

        :param n: The dimension of the embeddings.
        """
        self.n = n
        self.embedding_indices = dict()
        self.nr_embeddings = 0

    def _add_embedding(self, term):
        i = self.nr_embeddings
        self.nr_embeddings += 1
        if self.nr_embeddings > self.n:
            raise Exception("Number of one-hot encodings exceeds dimensions.")
        self.embedding_indices[term] = i
        return i

    def get_embedding(self, term):
        term = str(term)
        try:
            i = self.embedding_indices[term]
        except KeyError:
            i = self._add_embedding(term)
        tensor = torch.zeros((1, self.n))
        tensor[0, i] = 1.0
        return tensor


class Embedding(TermEmbedder):
    """
    Embeds terms with learnable embeddings. The memory for the embeddings is allocated in blocks.
    """

    def __init__(self, n: int, block_size: int = 32, renormalize: bool = False):
        """

        :param n: The dimension of the embeddings.
        :param block_size: How many embeddings are allocated per block.
        :param renormalize: If true, then each embedding is normalized after the update step.
        """
        self.embedding_indices = dict()
        self.keys = []
        self.optimizer = None
        self.renormalize = renormalize
        self.n = n
        self.block_size = block_size
        self.embeddings = [self._get_new_embeddings()]
        self.nr_embeddings = 0

    def _get_new_embeddings(self):
        return torch.nn.Embedding(self.block_size, self.n)

    def _grow(self):
        self.embeddings.append(self._get_new_embeddings())
        self.optimizer.add_param_group({"params": self.embeddings[-1].parameters()})

    def _add_embedding(self, term):
        i = self.nr_embeddings
        self.nr_embeddings += 1

        if i // self.block_size >= len(self.embeddings):
            self._grow()
        self.embedding_indices[term] = i
        self.keys.append(term)
        return i

    def get_embedding(self, term):
        term = str(term)
        try:
            i = self.embedding_indices[term]

        except KeyError:
            i = self._add_embedding(term)
        return self.embeddings[i // self.block_size](
            torch.LongTensor([i % self.block_size])
        )

    def parameters(self) -> List[torch.nn.Parameter]:
        """

        :return: A list of all parameters of the embeddings.
        """
        return [
            param for embedding in self.embeddings for param in embedding.parameters()
        ]

    def zero_grad(self):
        self.optimizer.zero_grad()

    def step(self):
        self.optimizer.step()
        if self.renormalize:
            for embed in self.embeddings:
                for j in range(embed.weight.shape[0]):
                    embed.weight[j, :].data /= 0.1 * embed.weight[j, :].detach().norm()

    def get_all_embeddings(self, n: Optional[int] = None) -> torch.Tensor:
        """

        :param n: Optional int. Limits the returned embeddings to the first n.
        :return: A tensor containing the embeddings.
        """
        if n is None:
            n = self.nr_embeddings
        return torch.cat([embedding.weight for embedding in self.embeddings])[:n]

    def get_embeddings_by_label(self, labels: Collection[Term]) -> torch.Tensor:
        """
        :param labels: A list of terms.
        :return: Returns the tensor of the embeddings of the terms in label in order.
        """
        return torch.cat([self.get_embedding(label) for label in labels])



from collections import defaultdict
from typing import Tuple, List

import torchvision.transforms as transforms
from problog.logic import Term, Constant, list2term

from deepproblog.dataset import ImageDataset
from deepproblog.query import Query

transform = transforms.Compose(
    [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
)

Card = List[str]
Hand = Tuple[str, str]


class PokerSeparate(ImageDataset):
    def __getitem__(self, i):
        return super().__getitem__("{}_{}".format(*i))

    def __len__(self):
        return len(self.data)

    def to_query(self, i):
        cards = self.data[i]
        cards, h1, h2, outcome, p = cards[:5], cards[5], cards[6], cards[7], cards[8]
        sub_images = [Term("card_{}".format(x)) for x in range(4)]
        images = [
            Term("tensor", Term(self.dataset, Constant(i), Constant(x)))
            for x in range(4)
        ]
        sub = {sub_images[i]: images[i] for i in range(4)}
        if self.extra_supervision and i % 5 == 0:
            return Query(
                Term(
                    "game",
                    list2term(sub_images),
                    Term("card", Term(cards[-1][0])),
                    Term(outcome),
                ),
                sub,
                p,
            )
        else:
            return Query(Term("game", list2term(sub_images), Term(outcome)), sub, p)

    def __init__(
        self, dataset, extra_supervision=False, probs=(0.25, 0.25, 0.25, 0.25)
    ):
        super().__init__("data/images/{}".format(dataset), transform=transform)
        self.data: List[
            Tuple[Card, Card, Card, Card, Card, Hand, Hand, str, float]
        ] = []
        self.suits = False
        self.extra_supervision = extra_supervision
        self.dataset = dataset
        count = defaultdict(int)
        with open("data/labels/{}.csv".format(dataset)) as f:
            for line in f:
                cards: List[Card] = [c.split(" of ") for c in line.strip().split(",")]
                count[cards[-1][0]] += 1
                h1, h2 = (
                    best_hand(cards[:2] + [cards[-1]], self.suits),
                    best_hand(cards[2:], self.suits),
                )
                result = get_outcome(h1, h2)
                result_probabilities = get_probabilities(
                    cards[:2], cards[2:4], self.suits, probs
                )
                p = result_probabilities[result]
                self.data.append(tuple(cards + [h1, h2, result, p]))
        print(dataset)
        for c in count:
            print("\t", c, count[c] / len(self.data))


values = ["jack", "queen", "king", "ace"]
suits = ["hearts", "clubs", "diamonds", "spades"]
types = ["high", "pair", "threeofakind", "straight", "flush"]
subranks = {v: values.index(v) for v in values}
subranks.update({"low": 0, "high": 1, None: 0})


def is_flush(hand):
    suit = hand[0][1]
    for _, s in hand:
        if s != suit:
            return False
    return "flush", None


def is_straight(hand):
    sorted_hand = sorted(hand, key=lambda x: values.index(x[0]))
    sorted_hand = list(x[0] for x in sorted_hand)
    if sorted_hand == ["jack", "queen", "king"]:
        return "straight", "low"
    if sorted_hand == ["queen", "king", "ace"]:
        return "straight", "high"
    return False


def three_of_a_kind(hand):
    value = hand[0][0]
    for v, _ in hand:
        if v != value:
            return False
    return "threeofakind", value


def two_of_a_kind(hand):
    sorted_hand = sorted(hand, key=lambda x: values.index(x[0]))
    sorted_hand = list(x[0] for x in sorted_hand)
    value = sorted_hand[1]
    if sorted_hand[0] == value or sorted_hand[2] == value:
        return "pair", value
    return False


def high_card(hand):
    sorted_hand = sorted(hand, key=lambda x: values.index(x[0]))
    sorted_hand = list(x[0] for x in sorted_hand)
    return "high", sorted_hand[-1]


def best_hand(hand, with_suits):
    if with_suits:
        possibilities = [
            is_flush,
            is_straight,
            three_of_a_kind,
            two_of_a_kind,
            high_card,
        ]
    else:
        possibilities = [is_straight, three_of_a_kind, two_of_a_kind, high_card]
    for t in possibilities:
        result = t(hand)
        if result:
            return result
    return None


def get_outcome(t1, t2):
    i1, i2 = types.index(t1[0]), types.index(t2[0])
    if i1 > i2:
        return "win"
    if i1 < i2:
        return "loss"
    i1, i2 = subranks[t1[1]], subranks[t2[1]]
    if i1 > i2:
        return "win"
    if i1 < i2:
        return "loss"
    return "draw"


def hand_to_term(hand):
    if hand[1] is None:
        return Term(hand[0])
    else:
        return Term(hand[0], Term(hand[1]))


def get_probabilities(c1, c2, use_suits, p):
    outcomes = {"win": 0.0, "loss": 0.0, "draw": 0.0}
    for c in [(v, s) for v in values for s in suits]:
        h1, h2 = best_hand(c1 + [c], use_suits), best_hand(c2 + [c], use_suits)
        result = get_outcome(h1, h2)
        outcomes[result] += p[values.index(c[0])] * 0.25
    return outcomes



import random


def random_list(length):
    l = []
    for i in range(length):
        l.append(random.randint(0, 9))

    return l


def next_example(length):
    l1 = random_list(length)
    l2 = sorted(l1)
    return l1, l2


def generate_examples(N, length, out_file):
    with open(out_file, "w") as f:
        for i in range(N):
            l1, l2 = next_example(length)
            f.write(
                "forth_sort([{}],[{}]).\n".format(
                    ",".join([str(l) for l in l1]), ",".join([str(l) for l in l2])
                )
            )


for i in [5, 6, 7, 8]:
    for j in [8, 64]:
        generate_examples(256, i, "train{}_test{}_train.txt".format(i, j))
        generate_examples(32, j, "train{}_test{}_test.txt".format(i, j))
        generate_examples(32, j, "train{}_test{}_dev.txt".format(i, j))



import pytest
from deepproblog.engines import ApproximateEngine

import pytest
if ApproximateEngine is None:
    pytest.skip('Approximate Inference is not available as PySDD is not installed', allow_module_level=True)

from deepproblog.model import Model
from deepproblog.query import Query
from problog.logic import Term, Constant, Var
from deepproblog.solver import SolverException


def _create_model(program, cache=False) -> Model:
    """Setup code: Load a program minimally"""
    model = Model(program, [], load=False)
    engine = ApproximateEngine(model, 10, ApproximateEngine.geometric_mean)
    model.set_engine(engine, cache=cache)
    return model


def test_cache_error():
    with pytest.raises(SolverException):
        _create_model("", cache=True)


def test_ad():
    program = """
0.5 :: a; 0.5 :: b.
c :- a,b.
d :- a,a.
    """
    model = _create_model(program)
    q = Query(Term("c"))
    r = model.solve([q])[0].result[Term("c")]
    assert pytest.approx(0.0) == r
    q = Query(Term("d"))
    r = model.solve([q])[0].result[Term("d")]
    assert pytest.approx(0.5) == r


def test_ad2():
    program = """
0.5 :: a; 0.5 :: b.
0.5 :: a; 0.5 :: b.
c :- a,b.
d :- a,a.
    """
    model = _create_model(program)
    q = Query(Term("c"))
    r = model.solve([q])[0].result[Term("c")]
    assert pytest.approx(0.5) == r
    q = Query(Term("d"))
    r = model.solve([q])[0].result[Term("d")]
    assert pytest.approx(0.75) == r


def test_ad3():
    program = """
0.5 :: a(X); 0.5 :: b(X).
c :- a(1),b(1).
d :- a(1),b(2).
    """
    model = _create_model(program)
    q = Query(Term("c"))
    r = model.solve([q])[0].result[Term("c")]
    assert r == pytest.approx(0.0)
    q = Query(Term("d"))
    r = model.solve([q])[0].result[Term("d")]
    assert r == pytest.approx(0.25)


def test_fact():
    program = """
0.8::d(_).
e :- d(1),d(2).
f :- d(1),e.
    """
    model = _create_model(program)
    q = Query(Term("e"))
    r = model.solve([q])[0].result
    assert pytest.approx(0.64) == r[Term("e")]
    q = Query(Term("f"))
    r = model.solve([q])[0].result
    assert pytest.approx(0.64) == r[Term("f")]


def test_fact2():
    program = """
0.6::a(_).
0.5::a(_).
c :- a(1),a(2).
d :- a(1),a(1).
    """
    model = _create_model(program)
    q = Query(Term("c"))
    r = model.solve([q])[0].result[Term("c")]
    assert pytest.approx(0.64) == r
    q = Query(Term("d"))
    r = model.solve([q])[0].result[Term("d")]
    assert pytest.approx(0.8) == r


def test_foreign():
    program = """
a(T,Out) :- evaluate(T, Out).
    """

    def evaluate(x):
        print("Inside evaluate", x)
        return Constant(int(x) ** 2)

    model = _create_model(program)
    model.register_foreign(evaluate, "evaluate", 1, 1)
    q = Query(Term("a", Constant(2), Var("X")))
    r = model.solve([q])[0].result[Term("a", Constant(2), Constant(4))]
    assert pytest.approx(1.0) == r
    q = Query(Term("a", Constant(2), Constant(4)))
    r = model.solve([q])[0].result[Term("a", Constant(2), Constant(4))]
    assert pytest.approx(1.0) == r
    q = Query(Term("a", Constant(2), Constant(5)))
    r = model.solve([q])[0]
    assert len(r.result) == 0


def test_foreign_text():
    program = """
a(Text,I,Word) :- ith_word(Text,I,Word).
    """

    def ith_word(text, i):
        text = text.value
        word = text.split(" ")[int(i)]
        return Constant(word)

    model = _create_model(program)
    model.register_foreign(ith_word, "ith_word", 2, 1)
    queries = [
        Query(Term("a", Constant("one two three"), Constant(x), Var("X")))
        for x in range(3)
    ]
    r = model.solve(queries)
    assert (
        pytest.approx(1.0)
        == r[0].result[
            Term("a", Constant("one two three"), Constant(0), Constant("one"))
        ]
    )
    assert (
        pytest.approx(1.0)
        == r[1].result[
            Term("a", Constant("one two three"), Constant(1), Constant("two"))
        ]
    )
    assert (
        pytest.approx(1.0)
        == r[2].result[
            Term("a", Constant("one two three"), Constant(2), Constant("three"))
        ]
    )


def test_assignment():
    program = """
a(X) :- X=3.
    """

    def evaluate(x):
        print("Inside evaluate", x)
        return Constant(x ** 2)

    model = _create_model(program)
    model.register_foreign(evaluate, "evaluate", 1, 1)
    q = Query(Term("a", Var("X")))
    r = model.solve([q])[0].result[Term("a", Constant(3))]
    assert pytest.approx(1.0) == r



import torch

from deepproblog.dataset import DataLoader
from deepproblog.engines import ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.Poker import PokerSeparate
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.optimizer import SGD
from deepproblog.train import train_model
from deepproblog.utils.standard_networks import smallnet

batch_size = 50
datasets = {
    "unfair": PokerSeparate(
        "unfair", probs=[0.2, 0.4, 0.15, 0.25], extra_supervision=True
    ),
    "fair_test": PokerSeparate("fair_test"),
}

dataset = "unfair"
net = Network(
    smallnet(pretrained=True, num_classes=4, size=(100, 150)), "net1", batching=True
)
net.optimizer = torch.optim.Adam(net.parameters(), lr=1e-4)
loader = DataLoader(datasets[dataset], batch_size)

model = Model("model.pl", [net])
model.set_engine(ExactEngine(model), cache=True)
model.optimizer = SGD(model, 5e-2)
model.add_tensor_source(dataset, datasets[dataset])
model.add_tensor_source("fair_test", datasets["fair_test"])

train_obj = train_model(
    model,
    loader,
    10,
    loss_function_name="mse",
    log_iter=len(datasets["unfair"]) // batch_size,
    test_iter=5 * len(datasets["unfair"]) // batch_size,
    test=lambda x: [
        ("Accuracy", get_confusion_matrix(model, datasets["fair_test"]).accuracy())
    ],
    infoloss=0.5,
)  # ,

cm = get_confusion_matrix(model, datasets["fair_test"])



from typing import Callable, TYPE_CHECKING

import torch
from problog.logic import Term, Constant, is_list, term2list

from deepproblog.engines import Engine

if TYPE_CHECKING:
    from deepproblog.model import Model


def embed(engine: Engine, term: Term):
    embedding = engine.model.get_embedding(term)[0, :]
    return Term("tensor", Constant(engine.tensor_store.store(embedding)))


def to_tensor(model: "Model", a):
    if type(a) is Term:
        if is_list(a):
            a = term2list(a)
        else:
            return model.get_tensor(a)
    # elif type(a) is Functor:
    #     return engine.tensor_store[int(a.args[0])]
    if type(a) is list:
        out = [to_tensor(model, x) for x in a]
        return [x for x in out if x is not None]
    else:
        return float(a)


def tensor_wrapper(engine: Engine, func: Callable, *args):
    model = engine.model
    inputs = [to_tensor(model, a) for a in args]
    out = func(*inputs)
    return model.store_tensor(out)


def tensor_wrapper_nondet(engine: Engine, func: Callable, *args):
    model = engine.model
    inputs = [to_tensor(model, a) for a in args]
    out = func(*inputs)
    return out


def rbf(x, y):
    return torch.exp(-torch.norm(x - y, 2))


def add(x, y):
    return x + y


def mul(x, y):
    return x * y


def dot(x, y):
    return torch.dot(x, y)


def sigmoid(x):
    return torch.sigmoid(x)


def max(x):
    x = torch.stack(x, 0)
    x, _ = torch.max(x, 0)
    return x


def mean(x):
    x = torch.stack(x, 0)
    x = torch.mean(x, 0)
    return x


def one_hot(i, n):
    x = torch.zeros(int(n))
    x[int(i)] = 1.0
    return x


def cat(tensors):
    return torch.cat(tensors)


def stack(tensors):
    return torch.stack(tensors)


def list_to_tensor(list_term):
    return torch.tensor(list_term)


def tensor_index(tensor, index):
    index = [int(x) for x in index]
    return tensor[index]


def less_than(tensor1, tensor2):
    if tensor1 < tensor2:
        return [()]
    return []

def register_tensor_predicates(engine: Engine):
    engine.register_foreign(lambda *x: embed(engine, *x), "embed", 1, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, rbf, *x), "rbf", 2, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, add, *x), "add", 2, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, mul, *x), "mul", 2, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, dot, *x), "dot", 2, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, max, *x), "max", 1, 1)
    engine.register_foreign(
        lambda *x: tensor_wrapper(engine, sigmoid, *x), "sigmoid", 1, 1
    )
    engine.register_foreign(lambda *x: tensor_wrapper(engine, mean, *x), "mean", 1, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, stack, *x), "stack", 1, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, cat, *x), "cat", 1, 1)
    engine.register_foreign(
        lambda *x: tensor_wrapper(engine, one_hot, *x), "one_hot", 2, 1
    )
    engine.register_foreign(lambda *x: tensor_wrapper(engine, list_to_tensor, *x), "list_to_tensor", 1, 1)
    engine.register_foreign(lambda *x: tensor_wrapper(engine, tensor_index, *x), "tensor_index", 2, 1)
    engine.register_foreign_nondet(lambda *x: tensor_wrapper_nondet(engine, less_than, *x), "less_than", 2, 0)


import random

coin1 = [("heads", 0.5), ("tails", 0.5)]
coin2 = [("heads", 0.5), ("tails", 0.5)]


def sample(dist):
    r = random.random()
    for v, p in dist:
        r -= p
        if r < 0:
            return v
    return dist[-1][1]


def next_example():
    c1 = sample(coin1)
    c2 = sample(coin1)
    return c1, c2


def generate_examples(N, fname):
    lines = list()
    for i in range(N):
        c1, c2 = next_example()
        lines.append(", ".join([c1, c2]))
    with open(fname + ".csv", "w") as f:
        f.write("\n".join(lines))


random.seed(1234)

generate_examples(100, "label_data/train")

generate_examples(20, "label_data/test")



from typing import Optional, Union

from problog.logic import Term
from torch import Tensor

TensorStoreIndex = Union[int, Term]


class TensorStore(object):
    """
    A class that allows tensor to be stored and associated with a unique identifier.
    """

    def __init__(self):
        self.tensor_store = dict()
        self.i = 0

    def store(
        self, tensor: Tensor, i: Optional[TensorStoreIndex] = None
    ) -> TensorStoreIndex:
        """
        Stores a tensor.
        :param tensor: The tensor to store.
        :param i: If given, then tensor will be stored with index i, otherwise, a unique integer will be used.
        :return: The unique identifier of the stored tensor.
        """
        if i is None:
            i = self.i
            self.i += 1
        self.tensor_store[i] = tensor
        return i

    def __getitem__(self, item):
        return self.tensor_store[item]

    def __contains__(self, item):
        return item in self.tensor_store

    def clear(self):
        """
        Clear the tensor store.
        :return:
        """
        self.tensor_store.clear()
        self.i = 0



try:
    from deepproblog.engines.approximate_engine import ApproximateEngine
except ModuleNotFoundError:
    import warnings

    warnings.warn("ApproximateEngine is not available as PySwip could not be found")
    ApproximateEngine = None
from deepproblog.engines.engine import Engine
from deepproblog.engines.exact_engine import ExactEngine



from typing import List, Union

import numpy as np

from deepproblog.utils import TabularFormatter


class ConfusionMatrix(object):
    def __init__(self, classes: Union[int, List[str]] = 0):
        if isinstance(classes, int):
            self.n = classes
            self.classes: List[Union[str, int]] = list(range(self.n))
        else:
            self.classes: List[Union[str, int]] = classes
            self.n = len(classes)
        self.matrix = np.zeros((self.n, self.n), dtype=np.uint)

    def get_index(self, c: Union[str, int]):
        if c not in self.classes:
            self.grow(c)
        return self.classes.index(c)

    def grow(self, c: Union[str, int]):
        self.classes.append(c)
        self.n = len(self.classes)
        new_matrix = np.zeros((self.n, self.n), dtype=np.uint)
        new_matrix[0 : self.n - 1, 0 : self.n - 1] = self.matrix
        self.matrix = new_matrix

    def add_item(self, predicted: Union[str, int], actual: Union[str, int]):
        actual_i = self.get_index(actual)
        predicted_i = self.get_index(predicted)

        self.matrix[predicted_i, actual_i] += 1

    def __str__(self):
        formatter = TabularFormatter()
        data = [[""] * (self.n + 2), ["", ""] + self.classes]
        data[0][(self.n + 1) // 2 + 1] = "Actual"
        for row in range(self.n):
            data.append(
                [" ", self.classes[row]]
                + [str(self.matrix[row, col]) for col in range(self.n)]
            )
        data[len(data) // 2][0] = "Predicted"
        return formatter.format(data)

    def accuracy(self):
        correct = 0
        for i in range(self.n):
            correct += self.matrix[i, i]
        total = self.matrix.sum()
        acc = correct / total
        print("Accuracy: ", acc)
        return acc



import bisect
from os import PathLike
from typing import Dict, Any, List, Union, Iterable

import numpy as np

from deepproblog.utils import check_path, split


class Logger(object):
    def __init__(self):
        self.log_data: Dict[str, Dict[int, Any]] = dict()
        self.indices: List[int] = list()
        self.comments: List[str] = list()

    def clear(self):
        """Clear history, needed when dealing with resetting training"""
        self.log_data.clear()
        self.indices.clear()
        self.comments.clear()

    def log(self, name: str, index: int, value: Any):
        if name not in self.log_data:
            self.log_data[name] = dict()
        i = bisect.bisect_left(self.indices, index)
        if i >= len(self.indices) or self.indices[i] != index:
            self.indices.insert(i, index)
        self.log_data[name][index] = value

    def log_list(self, i: int, data: Iterable):
        if data is not None:
            for e in data:
                self.log(e[0], i, e[1])

    def comment(self, comment: str):
        self.comments += comment.split("\n")

    def get_attribute(self, attribute: str, include_indices=False):
        sorted_attribute = []
        if attribute not in self.log_data:
            return []
        indices = []
        for i in self.indices:
            if i in self.log_data[attribute]:
                indices.append(i)
                sorted_attribute.append(self.log_data[attribute][i])
        if include_indices:
            return indices, sorted_attribute
        return sorted_attribute

    def get_union(self, att1: str, att2: str):
        indices = []
        sorted_attribute = []
        for i in self.indices:
            if i in self.log_data[att1] and i in self.log_data[att2]:
                indices.append(i)
                sorted_attribute.append(
                    (self.log_data[att1][i], self.log_data[att2][i])
                )

        return indices, sorted_attribute

    def __getitem__(self, item: str):
        return self.get_attribute(item, True)

    def __str__(self):
        lines = ["#{}".format(comment) for comment in self.comments]
        columns = list(self.log_data.keys())
        lines += ["i," + ",".join(columns)]
        for i in self.indices:
            row = [str(i)]
            for c in columns:
                row.append(str(self.log_data[c].get(i, "")))
            lines.append(",".join(row))
        return "\n".join(lines)

    def has_attribute(self, attribute):
        return attribute in self.log_data

    def write_to_file(self, name: str):
        filename = name + ".log"
        check_path(filename)
        with open(filename, "w") as f:
            f.write(str(self))

    def read_from_file(self, filename: Union[PathLike[str], str]):
        headers = None
        with open(filename) as f:
            for line in f:
                line = line.strip()
                if line[0] == "#":
                    self.comments.append(line[1:])
                    continue
                if headers is None:
                    headers = split(line, ",")
                    continue
                data = line.split(",")
                index = 0
                for i, h in enumerate(headers):
                    if h == "i":
                        index = int(data[i])
                        continue
                    if len(data[i]) > 0:
                        self.log(h, index, float(data[i]))


def equalize(data):  # Makes all sets of data contain same indices
    indices = set()
    for d in data:
        for i in d[0]:
            indices.add(i)
    indices = sorted(indices)
    new_data = []
    for d_ind, d_data in data:
        new_d = []
        prev_value = 0
        for i in indices:
            try:
                prev_value = d_data[d_ind.index(i)]
            except ValueError:
                pass
            new_d.append(prev_value)
        new_data.append(new_d)
    return indices, new_data


def aggregate(data, percentiles):
    matrix = np.zeros((len(data[0]), len(data)))
    for i, d in enumerate(data):
        matrix[:, i] = d
    stats = np.percentile(matrix, percentiles, 1)
    return stats


def aggregate_loggers(loggers, x, y, percentiles=(0.5,)):
    x_data = [logger[x] for logger in loggers]
    y_data = [logger[y] for logger in loggers]
    data = [tuple(equalize([x_data[i], y_data[i]])[1]) for i in range(len(x_data))]
    t, data = equalize(data)
    return t, aggregate(data, percentiles)



import pytest

from deepproblog.engines import ExactEngine, ApproximateEngine
from deepproblog.model import Model
from deepproblog.query import Query
from deepproblog.utils import parse

_simple_program = """
0.5 :: a.

equal(X,X).

0.1 :: multiple_answers(dummy1).
0.2 :: multiple_answers(dummy2). 
"""

@pytest.fixture(
    params=[
        {
            "name": "approximate",
            "engine_factory": lambda model: ApproximateEngine(model, 2, ApproximateEngine.geometric_mean),
            "cache": False,
        },
        {"name": "no_cache", "engine_factory": lambda model: ExactEngine(model), "cache": False},
        {"name": "cache", "engine_factory": lambda model: ExactEngine(model), "cache": True},
    ]
)
def model(request) -> Model:
    if ApproximateEngine is None and request.param["name"] == "approximate":
        pytest.skip(reason='ApproximateEngine is not available')
    """Simple fixture creating both the approximate and the exact engine"""
    model = Model(_simple_program, [], load=False)
    engine = request.param["engine_factory"](model)
    model.set_engine(engine, cache=request.param["cache"])
    return model


def test_model_basics(model):
    # These should be set up after running the fixture.
    assert model.solver is not None
    assert model.program is not None


def test_solve(model):
    q1 = Query(parse("equal(dummy,dummy)."))
    q2 = Query(parse("equal(dummy,notequal)."))
    q3 = Query(parse("multiple_answers(X)."))
    q4 = Query(parse("a."))
    results = model.solve([q1, q2, q3, q4])
    assert len(results) == 4
    # Q1:
    assert len(results[0].result) == 1  # Provable
    assert q1.query in results[0].result
    assert pytest.approx(1.0) == results[0].result[q1.query]
    # Q2:
    r2 = results[1].result
    assert len(r2) == 0 or (
        len(r2) == 1 and r2[q2.query] == pytest.approx(0.0)
    )  # Unprovable
    # Q3:
    r3 = results[2].result
    assert len(r3) == 2
    assert r3[parse("multiple_answers(dummy1).")] == pytest.approx(0.1)
    assert r3[parse("multiple_answers(dummy2).")] == pytest.approx(0.2)
    # Q4:
    r4 = results[3].result
    assert len(r4) == 1
    assert r4[q4.query] == pytest.approx(0.5)



import itertools
import json
import random
from pathlib import Path
from typing import Callable, List, Iterable, Tuple

import torchvision
import torchvision.transforms as transforms
from problog.logic import Term, list2term, Constant
from torch.utils.data import Dataset as TorchDataset

from deepproblog.dataset import Dataset
from deepproblog.query import Query

_DATA_ROOT = Path(__file__).parent

transform = transforms.Compose(
    [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
)

datasets = {
    "train": torchvision.datasets.MNIST(
        root=str(_DATA_ROOT), train=True, download=True, transform=transform
    ),
    "test": torchvision.datasets.MNIST(
        root=str(_DATA_ROOT), train=False, download=True, transform=transform
    ),
}


def digits_to_number(digits: Iterable[int]) -> int:
    number = 0
    for d in digits:
        number *= 10
        number += d
    return number


class MNIST_Images(object):
    def __init__(self, subset):
        self.subset = subset

    def __getitem__(self, item):
        return datasets[self.subset][int(item[0])][0]


MNIST_train = MNIST_Images("train")
MNIST_test = MNIST_Images("test")


class MNIST(Dataset):
    def __len__(self):
        return len(self.data)

    def to_query(self, i):
        l = Constant(self.data[i][1])
        return Query(
            Term("digit", Term("tensor", Term(self.dataset, Term("a"))), l),
            substitution={Term("a"): Constant(i)},
        )

    def __init__(self, dataset):
        self.dataset = dataset
        self.data = datasets[dataset]


def addition(n: int, dataset: str, seed=None):
    """Returns a dataset for binary addition"""
    return MNISTOperator(
        dataset_name=dataset,
        function_name="addition" if n == 1 else "multi_addition",
        operator=sum,
        size=n,
        arity=2,
        seed=seed,
    )


class MNISTOperator(Dataset, TorchDataset):
    def __getitem__(self, index: int) -> Tuple[list, list, int]:
        l1, l2 = self.data[index]
        label = self._get_label(index)
        l1 = [self.dataset[x][0] for x in l1]
        l2 = [self.dataset[x][0] for x in l2]
        return l1, l2, label

    def __init__(
        self,
        dataset_name: str,
        function_name: str,
        operator: Callable[[List[int]], int],
        size=1,
        arity=2,
        seed=None,
    ):
        """Generic dataset for operator(img, img) style datasets.

        :param dataset_name: Dataset to use (train, val, test)
        :param function_name: Name of Problog function to query.
        :param operator: Operator to generate correct examples
        :param size: Size of numbers (number of digits)
        :param arity: Number of arguments for the operator
        :param seed: Seed for RNG
        """
        super(MNISTOperator, self).__init__()
        assert size >= 1
        assert arity >= 1
        self.dataset_name = dataset_name
        self.dataset = datasets[self.dataset_name]
        self.function_name = function_name
        self.operator = operator
        self.size = size
        self.arity = arity
        self.seed = seed
        mnist_indices = list(range(len(self.dataset)))
        if seed is not None:
            rng = random.Random(seed)
            rng.shuffle(mnist_indices)
        dataset_iter = iter(mnist_indices)
        # Build list of examples (mnist indices)
        self.data = []
        try:
            while dataset_iter:
                self.data.append(
                    [
                        [next(dataset_iter) for _ in range(self.size)]
                        for _ in range(self.arity)
                    ]
                )
        except StopIteration:
            pass

    def to_file_repr(self, i):
        """Old file represenation dump. Not a very clear format as multi-digit arguments are not separated"""
        return f"{tuple(itertools.chain(*self.data[i]))}\t{self._get_label(i)}"

    def to_json(self):
        """
        Convert to JSON, for easy comparisons with other systems.

        Format is [EXAMPLE, ...]
        EXAMPLE :- [ARGS, expected_result]
        ARGS :- [MULTI_DIGIT_NUMBER, ...]
        MULTI_DIGIT_NUMBER :- [mnist_img_id, ...]
        """
        data = [(self.data[i], self._get_label(i)) for i in range(len(self))]
        return json.dumps(data)

    def to_query(self, i: int) -> Query:
        """Generate queries"""
        mnist_indices = self.data[i]
        expected_result = self._get_label(i)

        # Build substitution dictionary for the arguments
        subs = dict()
        var_names = []
        for i in range(self.arity):
            inner_vars = []
            for j in range(self.size):
                t = Term(f"p{i}_{j}")
                subs[t] = Term(
                    "tensor",
                    Term(
                        self.dataset_name,
                        Constant(mnist_indices[i][j]),
                    ),
                )
                inner_vars.append(t)
            var_names.append(inner_vars)

        # Build query
        if self.size == 1:
            return Query(
                Term(
                    self.function_name,
                    *(e[0] for e in var_names),
                    Constant(expected_result),
                ),
                subs,
            )
        else:
            return Query(
                Term(
                    self.function_name,
                    *(list2term(e) for e in var_names),
                    Constant(expected_result),
                ),
                subs,
            )

    def _get_label(self, i: int):
        mnist_indices = self.data[i]
        # Figure out what the ground truth is, first map each parameter to the value:
        ground_truth = [
            digits_to_number(self.dataset[j][1] for j in i) for i in mnist_indices
        ]
        # Then compute the expected value:
        expected_result = self.operator(ground_truth)
        return expected_result

    def __len__(self):
        return len(self.data)



from typing import Dict, Union

import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
from problog.logic import Term


class MLP(nn.Module):
    def __init__(self, *sizes, activation=nn.ReLU, softmax=True, batch=True):
        super(MLP, self).__init__()
        layers = []
        self.batch = batch
        for i in range(len(sizes) - 2):
            layers.append(nn.Linear(sizes[i], sizes[i + 1]))
            layers.append(activation())
        layers.append(nn.Linear(sizes[-2], sizes[-1]))
        if softmax:
            layers.append(nn.Softmax(-1))
        self.nn = nn.Sequential(*layers)

    def forward(self, x):
        if not self.batch:
            x = x.unsqueeze(0)
        x = self.nn(x)
        return x


model_urls = {
    "alexnet": "https://download.pytorch.org/models/alexnet-owt-4df8aa71.pth",
}


class DummyNet(nn.Module):
    def __init__(self, values: Dict[Union[Term, tuple[Term, ...]], Union[list, torch.Tensor]]):
        super().__init__()
        self.values = values

    def forward(self, *x: Term):
        if len(x) == 1:
            output = self.values[x[0]]
        else:
            output = self.values[x]
        return torch.tensor(output, requires_grad=True)


class DummyTensorNet(nn.Module):
    def __init__(self, batching=False):
        super().__init__()
        self.batching = batching

    def forward(self, *x: torch.Tensor):
        if self.batching:
            return torch.stack([torch.tensor(y, requires_grad=True) for y in x], dim=0)
        else:
            return torch.tensor(x, requires_grad=True)


class SmallNet(nn.Module):
    def __init__(self, num_classes=1000, size=None):
        super(SmallNet, self).__init__()
        self.final = nn.Sigmoid() if num_classes == 1 else nn.Softmax(1)
        self.num_classes = num_classes
        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(64, 192, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
            nn.Conv2d(192, 384, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(384, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=3, stride=2),
        )
        self.N = 2304
        if size is not None:
            inputs = torch.empty(1, 3, *size)
            out = self.features(inputs)
            self.N = torch.numel(out)
        self.classifier = nn.Sequential(
            nn.Linear(self.N, 100),
            nn.ReLU(inplace=True),
            nn.Linear(100, self.num_classes),
            self.final,
        )

    def forward(self, x):
        x = self.features(x)
        x = x.view(-1, self.N)
        x = self.classifier(x)
        return x


def smallnet(pretrained=False, model=None, **kwargs):
    if model is None:
        model = SmallNet(**kwargs)
    if pretrained:
        model.load_state_dict(model_zoo.load_url(model_urls["alexnet"]), strict=False)
    return model



from typing import TYPE_CHECKING, Sequence

import torch
from problog.formula import LogicFormula
from problog.logic import Term, Constant
from problog.program import LogicProgram

from deepproblog.query import Query
from deepproblog.tensor import TensorStore

if TYPE_CHECKING:
    from deepproblog.model import Model


class Engine(object):
    """
    An asbtract engine base class.
    """

    def __init__(self, model: "Model"):
        """

        :param model: The model that this engine will solve queries for.
        """
        self.model = model
        self.tensor_store = TensorStore()

    def perform_count(self, queries: Sequence[Query], acs):
        pass

    def prepare(self, program: LogicProgram) -> LogicProgram:
        """
        Modifies the given program to a format suited for querying in this engine.
        :param program: The program to be modified
        :return: The modified program
        """
        raise NotImplementedError("prepare is an abstract method")

    def ground(self, query: Query, **kwargs) -> LogicFormula:
        """

        :param query: The query to ground.
        :param kwargs:
        :return: A logic formula representing the ground program.
        """
        raise NotImplementedError("ground is an abstract method")

    def register_foreign(
        self, func: callable, function_name: str, arity_in: int, arity_out: int
    ):
        """
        Makes a Python function available to the grounding engine.
        :param func: The Python function to be made available.
        :param function_name: The name of the predicate that will be used to address this function in logic.
        :param arity_in: The number of input arguments to func
        :param arity_out The number of return values of func
        :return:
        """
        raise NotImplementedError("register_foreign is an abstract method")

    def register_foreign_nondet(
        self, func: callable, function_name: str, arity_in: int, arity_out: int
    ):
        """
        Makes a Python function available to the grounding engine.
        :param func: The Python function to be made available.
        :param function_name: The name of the predicate that will be used to address this function in logic.
        :param arity_in: The number of input arguments to func
        :param arity_out The number of return values of func
        :return:
        """
        raise NotImplementedError("register_foreign_nondet is an abstract method")


    def get_tensor(self, tensor_term: Term):
        if tensor_term.functor == "tensor":
            return self.tensor_store[tensor_term.args[0]]
        if type(tensor_term) is Constant and type(tensor_term.functor) is float:
            return torch.FloatTensor([tensor_term.functor])
        return tensor_term

    def get_hyperparameters(self) -> dict:
        raise NotImplementedError("get_hyperparameters is an abstract method")

    def eval(self):
        """
        Set the engine to eval mode.
        :return:
        """
        pass

    def train(self):
        """
        Set th engine to train mode.
        :return:
        """
        pass

    @staticmethod
    def can_cache() -> bool:
        return False



from json import dumps

import torch

from deepproblog.dataset import DataLoader
from deepproblog.engines import ApproximateEngine, ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.MNIST.data import MNIST_train, MNIST_test, addition
from deepproblog.examples.MNIST.network import MNIST_Net
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model

method = "exact"
N = 1

name = "addition_{}_{}".format(method, N)

train_set = addition(N, "train")
test_set = addition(N, "test")

network = MNIST_Net()

pretrain = 0
if pretrain is not None and pretrain > 0:
    network.load_state_dict(torch.load("models/pretrained/all_{}.pth".format(pretrain)))
net = Network(network, "mnist_net", batching=True)
net.optimizer = torch.optim.Adam(network.parameters(), lr=1e-3)

model = Model("models/addition.pl", [net])
if method == "exact":
    model.set_engine(ExactEngine(model), cache=True)
elif method == "geometric_mean":
    model.set_engine(
        ApproximateEngine(model, 1, ApproximateEngine.geometric_mean, exploration=False)
    )

model.add_tensor_source("train", MNIST_train)
model.add_tensor_source("test", MNIST_test)

loader = DataLoader(train_set, 2, False)
train = train_model(model, loader, 1, log_iter=100, profile=0)
model.save_state("snapshot/" + name + ".pth")
train.logger.comment(dumps(model.get_hyperparameters()))
train.logger.comment(
    "Accuracy {}".format(get_confusion_matrix(model, test_set, verbose=1).accuracy())
)
train.logger.write_to_file("log/" + name)



from problog.logic import (
    Term,
    AnnotatedDisjunction,
    term2list,
    Clause,
    Or,
    Constant,
    And,
    list2term,
)
from problog.program import SimpleProgram
from pyswip import Variable, registerForeign

from deepproblog.engines.engine import Engine
from deepproblog.engines.prolog_engine import (
    PrologEngine,
    pyswip_to_term,
    term_to_pyswip,
)
from deepproblog.engines.prolog_engine.heuristics import (
    GeometricMean,
    PartialProbability,
    Heuristic,
    LearnedHeuristic,
)
from deepproblog.engines.prolog_engine.swi_program import SWIProgram
from deepproblog.tensor import TensorStore


def wrap_tensor(x, store: TensorStore):
    if type(x) is list:
        return list2term([wrap_tensor(e, store) for e in x])
    else:
        return Term("tensor", Constant(store.store(x)))


def unwrap_tensor(x, model):
    if x.functor == "tensor":
        return model.get_tensor(x)
    else:
        return x


class ApproximateEngine(Engine):
    geometric_mean = GeometricMean()
    ucs = PartialProbability()

    def __init__(
            self,
            model,
            k,
            heuristic: Heuristic,
            exploration=False,
            timeout=None,
            ignore_timeout=False,
    ):
        Engine.__init__(self, model)
        self.heuristic = heuristic
        self.k = k
        self.engine = PrologEngine(k, heuristic, exploration, timeout, ignore_timeout)
        heuristic.engine = self.engine

    def perform_count(self, batch, acs):
        if isinstance(self.heuristic, LearnedHeuristic):
            self.heuristic.count(batch, acs)

    def prepare(self, db):
        translated = SimpleProgram()

        for e in db:
            new_e = e
            if type(e) is Term or type(e) is Clause:
                body = None
                if type(e) is Clause:
                    body = e.body
                    e = e.head
                p = e.probability
                if p is not None and p.functor == "nn":
                    if len(p.args) == 4:
                        net, inputs, output, domain = p.args
                        heads = []
                        self.model.networks[str(net)].domain = term2list(domain, False)
                        for domain_n, domain_element in enumerate(
                                term2list(domain, False)
                        ):
                            head = e.with_probability(
                                p.with_args(net, inputs, Constant(domain_n))
                            )
                            head = head.apply_term({output: domain_element})
                            heads.append(head)
                        if type(e) is Clause:
                            new_e = AnnotatedDisjunction(heads, e.body)
                        else:
                            new_e = Or.from_list(heads)
                    elif len(p.args) == 3:
                        net, inputs, output = p.args
                        self.model.networks[str(net)].det = True
                        head = e.with_probability(None)
                        body2 = Term(
                            "extern", Term("{}_extern".format(net), inputs, output)
                        )
                        if body is not None:
                            body = And(body, body2)
                        else:
                            body = body2
                        new_e = Clause(head, body)
                    elif len(p.args) == 2:
                        # net = p.args[0]
                        new_e = e
                    else:
                        raise ValueError(
                            "A neural predicate with {} arguments is not supported.".format(
                                len(p.args)
                            )
                        )
            translated.add_clause(new_e)
        db = self.engine.prepare(translated)
        self.register_networks(db)
        self.register_parameter(db)

        def evaluate(network, inputs2):
            inputs3 = tuple(term2list(inputs2, False))
            probabilities = self.model.networks[network]([inputs3])
            self.model.networks[network].cache[inputs2] = probabilities[0]
            probabilities = probabilities.tolist()[0]
            return probabilities

        db.eval_mode = evaluate
        self.register_tensor(db)
        return db

    def ground(self, query, label=None, **kwargs):
        return self.engine.ground(
            self.model.solver.program, query.substitute().query, label=label, **kwargs
        )

    def get_network_func_3(self, network):
        def func(arguments, output_var):
            arguments, variables = pyswip_to_term(arguments, with_variables=True)
            net = self.model.networks[network]
            out = net([term2list(arguments, False)])[0]
            out = wrap_tensor(out, self.tensor_store)
            out = term_to_pyswip(out)
            output_var.unify(out)

        func.arity = 2
        return func

    def get_network_func_4(self, network):
        def func(inputs, output_var):
            inputs, variables = pyswip_to_term(inputs, with_variables=True)
            assert len(variables) == 0
            net = self.model.networks[network]
            probabilities = net([term2list(inputs, False)])
            probabilities = probabilities[0]
            output_var.unify([float(x) for x in probabilities])

        func.arity = 2
        return func

    def get_network_func_2(self, network):
        def func(inputs, output_var):
            inputs, variables = pyswip_to_term(inputs, with_variables=True)
            assert len(variables) == 0
            net = self.model.networks[network]
            probabilities = net([term2list(inputs, False)])
            probabilities = probabilities[0]
            output_var.unify(float(probabilities[0]))

        func.arity = 2
        return func

    def register_networks(self, db: SWIProgram):
        for network in self.model.networks:
            name = network + "_extern"
            if self.model.networks[network].domain is not None:
                func = self.get_network_func_4(network)
                db.registerForeign(func, name)
            elif self.model.networks[network].det:
                func = self.get_network_func_3(network)
                db.registerForeign(func, name)
            else:
                func = self.get_network_func_2(network)
                db.registerForeign(func, name)

    def register_parameter(self, db: SWIProgram):
        def get_parameter(parameter_id, p):
            probability = self.model.parameters[parameter_id]
            p.unify(probability)

        get_parameter.arity = 2
        db.registerForeign(get_parameter, "get_parameter")

    def register_tensor(self, db):
        def get_tensor_probability(parameter_id, p):
            probability = float(self.tensor_store[parameter_id])
            p.unify(probability)

        get_tensor_probability.arity = 2
        db.registerForeign(get_tensor_probability, "get_tensor_probability")

    @staticmethod
    def get_wrapped_func(func, arity_in, arity_out):
        def wrapped_func(*arguments):
            input_args, output_args = arguments[:arity_in], arguments[arity_in:]
            input_args = [pyswip_to_term(x) for x in input_args]
            result = func(*input_args)
            if type(result) is not func:
                result = (result,)
            result = [term_to_pyswip(r) for r in result]
            for o, r in zip(output_args, result):
                if type(o) is Variable:
                    o.unify(r)
                else:
                    if o != r:
                        return False

        return wrapped_func

    def register_foreign(self, func, function_name, arity_in, arity_out):
        wrapped_func = self.get_wrapped_func(func, arity_in, arity_out)
        wrapped_func.arity = arity_in + arity_out
        registerForeign(wrapped_func, function_name)
        builtin_name = "{}({})".format(
            function_name, ",".join(["_"] * (arity_in + arity_out))
        )
        list(
            self.engine.prolog.query(
                "assertz(allowed_builtin({}))".format(builtin_name)
            )
        )

    def register_foreign_nondet(self, func, function_name, arity_in, arity_out):
        #TODO Implement
        pass

    def get_hyperparameters(self) -> dict:
        parameters = {
            "type": "ApproximateEngine",
            "heuristic": self.heuristic.get_hyperparameters(),
            "k": self.k,
        }
        return parameters

    def eval(self):
        list(self.engine.prolog.query("set_flag(mode,eval)"))

    def train(self):
        list(self.engine.prolog.query("set_flag(mode,train)"))



from abc import ABC, abstractmethod
from typing import Optional, Dict, Union

import torch
from problog.evaluator import Semiring as ProbLogSemiring
from problog.formula import LogicFormula
from problog.logic import Term

from deepproblog.query import Query


class Semiring(ProbLogSemiring, ABC):
    """
    The semiring object defines the operations for the evaluation of arithmetic circuits.
    """

    def __init__(self, model, substitution, values):
        """
        :param model: The model in which the evaluation happens.
        :param substitution: The substitution to apply to the arithmetic circuit before evaluation.
        :param values: The output values of the neural network to use in the evaluation.
        """
        self.model = model
        self.eps = 1e-5
        self.values = values
        self.substitution = substitution

    @staticmethod
    @abstractmethod
    def cross_entropy(
        result: "Result",
        target: float,
        weight: float,
        q: Optional[Query] = None,
        eps: float = 1e-6,
    ):
        """
        Calculates the cross_entropy between the predicted and target probabilities.
        Also performs the backwards pass for the given result.
        :param result: The result to calculate loss on.
        :param target: The target probability.
        :param weight: The weight of this examplE. A float that is multiplied with the loss before backpropagation.
        :param q: If there's more than one query in result, calculate the loss for this query.
        :param eps: The epsilon used in the cross-entropy loss calculation.
        :return:
        """
        pass


class Result(object):
    """
    A class that contains the result and timing info for evaluating a query.
    """

    def __init__(
        self,
        result: Dict[Term, Union[float, torch.Tensor]],
        semiring: Semiring,
        ground_time: Optional[float] = None,
        compile_time: Optional[float] = None,
        eval_time: Optional[float] = None,
        proof: Optional[LogicFormula] = None,
    ):
        """Construct object

        :param result: Dictionary of results, the key is the term and the value is the probability.
        :param semiring: Semiring object in use
        :param ground_time:
        :param compile_time:
        :param eval_time:
        :param proof:

        Note! The term indexing the result object may not be the same as in your query. There are
        a few reasons for this:
        * Your query had substitutions, the term is going to be the substituted variant.
        * You have a non-ground query, your query could be partially ground (giving you multiple answers).
        """
        self.result = result
        self.semiring = semiring
        self.ground_time = ground_time
        self.compile_time = compile_time
        self.eval_time = eval_time
        self.proof = proof

    def __iter__(self):
        return iter(self.result.keys())

    def __len__(self):
        return len(self.result)

    def __repr__(self):
        return repr(self.result)



import signal
import time
from typing import List, Callable, Union

from deepproblog.dataset import DataLoader
from deepproblog.model import Model
from deepproblog.query import Query
from deepproblog.utils.logger import Logger
from deepproblog.utils.stop_condition import EpochStop
from deepproblog.utils.stop_condition import StopCondition


class TrainObject(object):
    """
    An object that performs the training of the model and keeps track of the state of the training.
    """

    def __init__(self, model: Model):
        self.model = model
        self.logger = Logger()
        self.accumulated_loss = 0
        self.i = 1
        self.start = 0
        self.prev_iter_time = 0
        self.epoch = 0
        self.previous_handler = None
        self.interrupt = False
        self.hooks = []
        self.timing = [0, 0, 0]

    def get_loss(self, batch: List[Query], backpropagate_loss: Callable) -> float:
        """
        Calculates and propagates the loss for a given batch of queries and loss function.
        :param batch: The batch of queries.
        :param backpropagate_loss:  The loss function. It should also perform the backpropagation.
        :return: The average loss over the batch
        """
        total_loss = 0
        result = self.model.solve(batch)
        for r in result:
            self.timing[0] += r.ground_time / len(batch)
            self.timing[1] += r.compile_time / len(batch)
            self.timing[2] += r.eval_time / len(batch)
        result = [
            (result[i], batch[i]) for i in range(len(batch)) if len(result[i]) > 0
        ]
        for r, q in result:
            total_loss += backpropagate_loss(
                r, q.p, weight=1 / len(result), q=q.substitute().query
            )
        return total_loss

    def get_loss_with_negatives(
        self, batch: List[Query], backpropagate_loss: Callable
    ) -> float:
        """
        Calculates and propagates the loss for a given batch of queries and loss function.
        This includes negative examples. Negative examples are found by using the query.replace_var method.
        :param batch: The batch of queries.
        :param backpropagate_loss:  The loss function. It should also perform the backpropagation.
        :return: The average loss over the batch
        """
        total_loss = 0

        result = self.model.solve([q.variable_output() for q in batch])
        result = [(result[i], batch[i]) for i in range(len(batch))]

        for r, q in result:
            expected = q.substitute().query
            try:
                total_loss += backpropagate_loss(
                    r, q.p, weight=1 / len(result), q=expected
                )
            except KeyError:
                self.get_loss([q], backpropagate_loss)
            neg_proofs = [x for x in r if x != expected]
            for neg in neg_proofs:
                # print('penalizing wrong answer {} vs {}'.format(q.substitute().query, k))
                total_loss += backpropagate_loss(
                    r, 0, weight=1 / (len(result) * len(neg_proofs)), q=neg
                )
        return total_loss

    def train(
        self,
        loader: DataLoader,
        stop_criterion: Union[int, StopCondition],
        verbose: int = 1,
        loss_function_name: str = "cross_entropy",
        with_negatives: bool = False,
        log_iter: int = 100,
        initial_test: bool = True,
        **kwargs
    ) -> Logger:

        self.previous_handler = signal.getsignal(signal.SIGINT)
        loss_function = getattr(self.model.solver.semiring, loss_function_name)

        self.accumulated_loss = 0
        self.timing = [0, 0, 0]
        self.epoch = 0
        self.start = time.time()
        self.prev_iter_time = time.time()
        epoch_size = len(loader)
        if "test" in kwargs and initial_test:
            value = kwargs["test"](self.model)
            self.logger.log_list(self.i, value)
            print("Test: ", value)

        if type(stop_criterion) is int:
            stop_criterion = EpochStop(stop_criterion)
        print("Training ", stop_criterion)

        while not (stop_criterion.is_stop(self) or self.interrupt):
            epoch_start = time.time()
            self.model.optimizer.step_epoch()
            if verbose and epoch_size > log_iter:
                print("Epoch", self.epoch + 1)
            for batch in loader:
                if self.interrupt:
                    break
                self.i += 1
                self.model.train()
                self.model.optimizer.zero_grad()
                if with_negatives:
                    loss = self.get_loss_with_negatives(batch, loss_function)
                else:
                    loss = self.get_loss(batch, loss_function)
                self.accumulated_loss += loss

                self.model.optimizer.step()
                self.log(verbose=verbose, log_iter=log_iter, **kwargs)
                for j, hook in self.hooks:
                    if self.i % j == 0:
                        hook(self)

                if stop_criterion.is_stop(self):
                    break
            if verbose and epoch_size > log_iter:
                print("Epoch time: ", time.time() - epoch_start)
            self.epoch += 1
        if "snapshot_name" in kwargs:
            filename = "{}_final.mdl".format(kwargs["snapshot_name"])
            print("Writing snapshot to " + filename)
            self.model.save_state(filename)

        signal.signal(signal.SIGINT, self.previous_handler)
        return self.logger

    def log(
        self, snapshot_iter=None, log_iter=100, test_iter=1000, verbose=1, **kwargs
    ):
        iter_time = time.time()

        if (
            "snapshot_name" in kwargs
            and snapshot_iter is not None
            and self.i % snapshot_iter == 0
        ):
            filename = "{}_iter_{}.mdl".format(kwargs["snapshot_name"], self.i)
            print("Writing snapshot to " + filename)
            self.model.save_state(filename)
        if verbose and self.i % log_iter == 0:
            print(
                "Iteration: ",
                self.i,
                "\ts:%.4f" % (iter_time - self.prev_iter_time),
                "\tAverage Loss: ",
                self.accumulated_loss / log_iter,
            )
            if len(self.model.parameters):
                print("\t".join(str(parameter) for parameter in self.model.parameters))
            self.logger.log("time", self.i, iter_time - self.start)
            self.logger.log("loss", self.i, self.accumulated_loss / log_iter)
            self.logger.log("ground_time", self.i, self.timing[0] / log_iter)
            self.logger.log("compile_time", self.i, self.timing[1] / log_iter)
            self.logger.log("eval_time", self.i, self.timing[2] / log_iter)
            # for k in self.model.parameters:
            #     self.logger.log(str(k), self.i, self.model.parameters[k])
            #     print(str(k), self.model.parameters[k])
            self.accumulated_loss = 0
            self.timing = [0, 0, 0]
            self.prev_iter_time = iter_time
        if "test" in kwargs and self.i % test_iter == 0:
            value = kwargs["test"](self.model)
            self.logger.log_list(self.i, value)
            print("Test: ", value)

    def write_to_file(self, *args, **kwargs):
        self.logger.write_to_file(*args, **kwargs)


def train_model(
    model: Model,
    loader: DataLoader,
    stop_condition: Union[int, StopCondition],
    **kwargs
) -> TrainObject:
    train_object = TrainObject(model)
    train_object.train(loader, stop_condition, **kwargs)
    return train_object



from collections import defaultdict
from statistics import mean
from typing import Optional

from deepproblog.dataset import Dataset
from deepproblog.model import Model
from deepproblog.utils.confusion_matrix import ConfusionMatrix


def get_confusion_matrix(
    model: Model, dataset: Dataset, verbose: int = 0, eps: Optional[float] = None
) -> ConfusionMatrix:
    """

    :param model: The model to evaluate.
    :param dataset: The dataset to evaluate the model on.
    :param verbose: Set the verbosity. If verbose > 0, then print confusion matrix and accuracy.
    If verbose > 1, then print all wrong answers.
    :param eps: If set, then the answer will be treated as a float, and will be considered correct if
    the difference between the predicted and ground truth value is smaller than eps.
    :return: The confusion matrix when evaluating model on dataset.
    """
    confusion_matrix = ConfusionMatrix()
    model.eval()
    for i, gt_query in enumerate(dataset.to_queries()):
        test_query = gt_query.variable_output()
        answer = model.solve([test_query])[0]
        actual = str(gt_query.output_values()[0])
        if len(answer.result) == 0:
            predicted = "no_answer"
            if verbose > 1:
                print("no answer for query {}".format(gt_query))
        else:
            max_ans = max(answer.result, key=lambda x: answer.result[x])
            p = answer.result[max_ans]
            if eps is None:
                predicted = str(max_ans.args[gt_query.output_ind[0]])
            else:
                predicted = float(max_ans.args[gt_query.output_ind[0]])
                actual = float(gt_query.output_values()[0])
                if abs(actual - predicted) < eps:
                    predicted = actual
            if verbose > 1 and actual != predicted:
                print(
                    "{} {} vs {}::{} for query {}".format(
                        i, actual, p, predicted, test_query
                    )
                )
        confusion_matrix.add_item(predicted, actual)

    if verbose > 0:
        print(confusion_matrix)
        print("Accuracy", confusion_matrix.accuracy())

    return confusion_matrix


def get_fact_accuracy(
    model: Model,
    dataset: Dataset,
    verbose: int = 0,
    threshold: float = 0.5,
    empty_answer: str = "false",
) -> ConfusionMatrix:
    """

    :param model: The model to evaluate.
    :param dataset: The dataset to evaluate the model on.
    :param verbose: Set the verbosity. If verbose > 0, then print confusion matrix and accuracy.
    If verbose > 1, then print all wrong answers.
    :param threshold: If set, then the answer will be treated as a float, and will be considered correct if
    the difference between the predicted and ground truth value is smaller than eps.
    :param empty_answer: The category in the confusion matrix when no answer was returned. Default = "false"
    :return: The confusion matrix when evaluating model on dataset.
    """
    confusion_matrix = ConfusionMatrix()
    model.eval()
    probabilities = defaultdict(list)
    for i, query in enumerate(dataset.to_queries()):
        answer = model.solve([query])[0]
        if len(answer.result) == 0:
            predicted = empty_answer
            if verbose > 1:
                print("no answer for query {}".format(query))
        else:
            predicted_p = float(answer.result[query.substitute().query])
            predicted = "true" if predicted_p >= threshold else "false"
            probabilities[predicted].append(predicted_p)
        actual = "true" if query.p >= threshold else "false"
        confusion_matrix.add_item(predicted, actual)
    for c in probabilities:
        print("Average probability for class {}: {}".format(c, mean(probabilities[c])))
    if verbose > 0:
        print(confusion_matrix)
        print("Accuracy", confusion_matrix.accuracy())

    return confusion_matrix


# def multi_class_accuracy(model, data, test=False, verbose=1, inds=None):
#     if inds is None:
#         inds = [-1]
#     correct = 0
#     sub = None
#     N = 0
#     model.eval()
#     for batch in data:
#         N += len(batch)
#         query_batch = []
#         for q in batch:
#             if type(q) is tuple:
#                 q, sub = q
#             args = list(q.args)
#             for i in inds:
#                 args[i] = Var('X_{}'.format(i))
#             q = q(*args)
#             if sub is not None:
#                 q = (q, sub)
#             query_batch.append(q)
#
#         result, _ = model.solve(query_batch, test)
#         for i, out in enumerate(result):
#             out = max(out, key=lambda x: out[x][0])
#             if type(batch[i]) is tuple:
#                 res_out = batch[i][0]
#                 if test:
#                     res_out = res_out.apply_term(batch[i][1])
#             else:
#                 res_out = batch[i]
#             if out == res_out:
#                 correct += 1
#                 if verbose > 2:
#                     print('Correct', out)
#             else:
#                 if verbose > 1:
#                     print('Wrong', res_out, 'vs', out)
#     if verbose > 0:
#         print('Accuracy', correct / N)
#     return [('Accuracy', correct / N)]
#
#
# def fact_accuracy(model, data, test=False, threshold=0.5, verbose=1):
#     correct = 0
#     N = 0
#     model.eval_mode()
#     for batch in data:
#         N += len(batch)
#         result, _ = model.solve(batch, test)
#         for i, out in enumerate(result):
#             if type(batch[i]) is tuple:
#                 q = batch[i][0]
#             else:
#                 q = batch[i]
#             p = out[q][0]
#             if p >= threshold:
#                 correct += 1
#             if verbose > 1:
#                 print(p, q)
#     if verbose > 0:
#         print('Accuracy', correct / N)
#     return [('Accuracy', correct / N)]
#
#



import re

for train in [2, 4, 8]:
    for test in [8, 64]:
        for mode in ["train", "test", "dev"]:
            with open("train{}_test{}/{}.txt".format(train, test, mode)) as f:
                text = f.read()
            matches = re.findall("\[([0-9 ]*)\]\t\[([0-9 ]*) (\d) (\d*)\]", text)
            text = list()
            for match in matches:
                res = match[0].strip().split(" ")
                digits = match[1].strip().split(" ")
                carry = [match[2]]
                length = int(match[3])
                digit1 = list()
                digit2 = list()
                for i in range(0, len(digits), 2):
                    digit1.append(digits[i])
                    digit2.append(digits[i + 1])
                text.append(
                    "add([{}],[{}],{},[{}]).".format(
                        *[",".join(l) for l in [digit1, digit2, carry, res]]
                    )
                )

            with open("train{}_test{}_{}.txt".format(train, test, mode), "w") as f:
                f.write("\n".join(text))



import os

import torchvision.transforms as transforms

from deepproblog.dataset import ImageDataset
from deepproblog.query import Query
from problog.logic import Term, Constant

path = os.path.dirname(os.path.abspath(__file__))

transform = transforms.Compose(
    [transforms.ToTensor(), transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))]
)


class Coins(ImageDataset):
    def __init__(
        self, subset,
    ):
        super().__init__("{}/image_data/{}/".format(path, subset), transform=transform)
        self.data = []
        self.subset = subset
        with open("{}/label_data/{}.csv".format(path, subset)) as f:
            for line in f:
                c1, c2 = [l.strip() for l in line.split(",")]
                outcome = "loss"
                if c1 == c2:
                    outcome = "win"
                self.data.append((c1, c2, outcome))

    def to_query(self, i):
        c1, c2, outcome = self.data[i]
        sub = {Term("a"): Term("tensor", Term(self.subset, Constant(i)))}
        return Query(Term("game", Term("a"), Term(outcome)), sub)

    def __len__(self):
        return len(self.data)


train_dataset = Coins("train")
test_dataset = Coins("test")



import random
import sys
from math import pi

import bpy
# noinspection PyUnresolvedReferences,PyUnresolvedReferences
from mathutils import Vector

argv = sys.argv
argv = argv[argv.index("--") + 1 :]

context = bpy.context
scene = context.scene
camera = bpy.data.objects["Camera"]
coin1 = bpy.data.objects["Coin1"]
coin2 = bpy.data.objects["Coin2"]
lamp = bpy.data.objects["Lamp"]

res = int(argv[2])
scene.render.resolution_x = res
scene.render.resolution_y = res

max_offset = 0.8


def render_csv(f_in, f_out):
    with open(f_in) as f:
        data = [l.strip().split(", ") for l in f]
    for i, (c1, c2) in enumerate(data):
        offset = [max_offset * (random.random() - 0.5) for i in range(2)] + [0]
        coin1.delta_location = Vector(offset)
        flip = c1 == "tails"
        coin1.delta_rotation_euler = [pi * flip, 0, random.random() * 2 * pi]

        offset = [max_offset * (random.random() - 0.5) for i in range(2)] + [0]
        coin2.delta_location = Vector(offset)
        flip = c2 == "tails"
        coin2.delta_rotation_euler = [pi * flip, 0, random.random() * 2 * pi]

        scene.camera = camera
        bpy.data.scenes["Scene"].render.filepath = f_out + "{}.png".format(i)
        bpy.ops.render.render(write_still=True)


render_csv(
    "{}/label_data/{}.csv".format(argv[0], argv[1]),
    "{}/image_data/{}/".format(argv[0], argv[1]),
)



carries = list()
results = list()

for n1 in range(0, 10):
    for n2 in range(0, 10):
        for carry in [0, 1]:
            ncarry = (n1 + n2 + carry) // 10
            res = (n1 + n2 + carry) % 10
            carries.append((n1, n2, carry, ncarry))
            results.append((n1, n2, carry, res))

carries = ["carry({},{},{},{}).".format(*c) for c in carries]
results = ["result({},{},{},{}).".format(*r) for r in results]

with open("tests.pl", "w") as f:
    f.write("\n".join(carries + results))



import os
import sys
from configparser import ConfigParser
from datetime import datetime
from itertools import islice
from pathlib import Path
from statistics import mean, stdev
from time import strftime
from typing import Union, Any, Dict

import problog
from problog.logic import Term

parser = problog.parser.PrologParser(problog.program.ExtendedPrologFactory())

cred = "\033[91m"
cend = "\033[0m"
cgreen = "\033[92m"


def log_exists(location: Union[str, os.PathLike], name: str):
    return Path(location).glob(name + "*")


def check_path(path: Union[str, os.PathLike]):
    path_dir = os.path.dirname(str(path))
    if not os.path.exists(path_dir):
        os.makedirs(path_dir)


def get_top_path(pattern: str, reverse=True):
    paths = sorted(Path(".").glob(pattern), reverse=reverse)
    if len(paths) > 0:
        return paths[0]
    return None


def split(text: str, splitchar: str, lb="(", rb=")"):
    depth = 0
    splits = [""]
    for c in text:
        if c in splitchar and depth == 0:
            splits.append("")
            continue
        if c in lb:
            depth += 1
        if c in rb:
            depth -= 1
        splits[-1] += c
    return splits


def count_parameters(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def nth(iterable, n, default=None):
    """Returns the nth item or a default value"""
    return next(islice(iterable, n, None), default)


def get_configuration(parameters: dict, i: int) -> dict:
    config = dict()
    for k in parameters:
        length = len(parameters[k])
        j, i = i % length, i // length
        config[k] = parameters[k][j]
    return config


def parse(string: str) -> Term:
    parsed = parser.parseString(string)
    for term in parsed:
        return term


class Table(object):
    class Dimension(object):
        def __init__(self, name):
            self.name = name
            self.type = "categorical"
            self.categories = set()

        def add_value(self, val):
            self.categories.add(val)

        def __repr__(self):
            return self.name

    def __init__(self, *dimensions):
        self.data = []
        self.dimensions = [Table.Dimension(d) for d in dimensions]

    def store(self, *args, **kwargs):
        if len(args) == 0:
            data = tuple(kwargs[str(d)] for d in self.dimensions)
        else:
            if len(args) != len(self.dimensions):
                raise ValueError(
                    "Data dimensions {} not equal to table dimensions {}".format(
                        len(args), len(self.dimensions)
                    )
                )
            data = args
        self.data.append(data)
        for i, d in enumerate(self.dimensions):
            d.add_value(data[i])

    def get_dimension_index(self, dim):
        for i, d in enumerate(self.dimensions):
            if hash(dim) == hash(d):
                return i
        raise ValueError("{} not in dimensions".format(dim))

    @staticmethod
    def aggregate(data):
        if data is None or len(data) == 0:
            return ""
        mu, sig = mean(data), stdev(data)
        return "{:.5f} ± {:.5f}".format(mu, sig)

    def format(self, x, y, val, **kwargs):
        categories = tuple(kwargs.items())
        categories = tuple((self.get_dimension_index(c[0]), c[1]) for c in categories)
        x = self.get_dimension_index(x)
        y = self.get_dimension_index(y)
        val = self.get_dimension_index(val)
        x_cats = list(self.dimensions[x].categories)
        y_cats = list(self.dimensions[y].categories)
        data = [[list() for _ in x_cats] for _ in y_cats]
        for d in self.data:
            j = x_cats.index(d[x])
            i = y_cats.index(d[y])
            correct_categories = True
            for k, v in categories:
                if d[k] != v:
                    correct_categories = False
                    break
            if correct_categories:
                data[i][j].append(d[val])
        data = [[self.aggregate(d) for d in row] for row in data]
        return TabularFormatter.format(data, x_cats, y_cats)


class TabularFormatter(object):
    @staticmethod
    def format(data, x=None, y=None):
        if y is not None:
            data = [[y[i]] + data[i] for i in range(len(data))]
            if x is not None:
                data = [[""] + x] + data
        else:
            if x is not None:
                data = x + data

        nr_columns = len(data[0])
        column_widths = [0] * nr_columns
        for row in data:
            for i, value in enumerate(row):
                column_widths[i] = max(column_widths[i], len(str(value)))

        padded_rows = [
            "\t".join(
                [
                    " " * (column_widths[i] - len(str(v))) + str(v)
                    for i, v in enumerate(row)
                ]
            )
            for row in data
        ]
        return "\n".join(padded_rows)


def format_time():
    return strftime("_%y%m%d_%H%M")


def format_time_precise():
    return datetime.utcnow().strftime("%y%m%d_%H%M%S%f")


class NoConfigException(Exception):
    def __str__(self):
        return "No config file specified as an argument."


def load_config(filename: str = None):
    """
    Loads a config file.
    :param filename: Filename of configuration file to load. If None, use the first commandline argument as filename.
    :return: None
    """
    try:
        if filename is None:
            filename = sys.argv[1]
        config = ConfigParser()
        config.read(filename)
        return config["Default"]
    except IndexError:
        raise NoConfigException()


def term2list2(term: Term):
    result = []
    while (
        not problog.logic.is_variable(term) and term.functor == "." and term.arity == 2
    ):
        result.append(term.args[0])
        term = term.args[1]
    if not term == problog.logic.Term("[]"):
        raise ValueError("Expected fixed list.")
    return result


def config_to_string(configuration: Dict[str, Any]) -> str:
    return "_".join(
        "{}_{}".format(parameter, configuration[parameter])
        for parameter in configuration
    )



import argparse
import os
import subprocess

parser = argparse.ArgumentParser(
    description="Render the image data for a given csv file."
)
parser.add_argument("set", nargs="+")
parser.add_argument("-b", "--blender_path", default=None)

if __name__ == "__main__":
    parsed = parser.parse_args()
    blender_path = parsed.blender_path
    if blender_path is None:
        blender_path = "blender"
    for s in parsed.set:
        print("Rendering ", s)

        path = os.path.dirname(os.path.abspath(__file__))
        res = 512
        subprocess.call(
            [
                blender_path,
                path + "/blender_files/scene.blend1",
                "-b",
                "-P",
                path + "/blender_files/render_script.py",
                "--",
                path,
                s,
                str(res),
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.STDOUT,
        )



import random
import subprocess
from pathlib import Path

import numpy as np
from PIL import Image

values = ["jack", "queen", "king", "ace"]
suits = ["hearts", "clubs", "diamonds", "spades"]


def render_svg(in_path, out_path, height):
    print("Rendering {} to {}", in_path, out_path)
    Path(out_path).parent.mkdir(parents=True, exist_ok=True)
    subprocess.run(
        [
            "inkscape",
            in_path,
            "-h",
            str(height),
            "--export-filename=" + out_path,
            "-b",
            '"#FFFFFF"',
        ]
    )


def get_images(img_h):
    images = {}
    for value in values:
        for suit in suits:
            path = Path("cards/rendered/{2}/{0}_{1}.png".format(value, suit, img_h))
            if not path.is_file():
                render_svg(
                    "cards/{}_{}.svg".format(value, suit),
                    "cards/rendered/{2}/{0}_{1}.png".format(value, suit, img_h),
                    img_h,
                )
            images[(value, suit)] = Image.open(path)
    return images


def render_all(games, img_h, path):
    imgs = get_images(img_h)
    path.mkdir(parents=True, exist_ok=True)
    for i, cards in enumerate(games):
        images = [apply_transformations(imgs[card]) for card in cards]
        for j, img in enumerate(images):
            img.save(path / "{}_{}.png".format(i, j))


def apply_transformations(image, max_rot=10, hue=13, sat=13, val=13):
    image = image.convert("HSV")
    data = np.array(image, dtype="int16")

    for i, v in enumerate([hue, sat, val]):
        data[:, :, i] += np.random.randint(-v, v, dtype="int16")
        data[:, :, 0] = data[:, :, 0] % 255
        data = data.clip(0, 255)
    image = Image.fromarray(data.astype("uint8"), mode="HSV")
    image = image.convert("RGB")

    w, h = image.width, image.height
    theta = random.uniform(-max_rot, max_rot)
    image = image.rotate(
        theta, resample=Image.BILINEAR, expand=True, fillcolor="#FFFFFF"
    )
    image = image.resize((w, h))

    return image


def read_file(path):
    games = []
    with open(path) as f:
        for line in f:
            cards = line.strip().split(",")[:4]
            cards = [tuple(c.split(" of ")) for c in cards]
            games.append(cards)
    return games


IMG_H = 150
for dataset in ["fair", "unfair"]:  # , 'more_jacks', 'more_spades','ace_spades']:
    render_all(
        read_file("labels/{}.csv".format(dataset)),
        IMG_H,
        Path("images/{}/".format(dataset)),
    )

for dataset in [
    "fair_test",
    "unfair_test",
]:  # , 'more_jacks_test', 'more_spades_test','ace_spades_test']:
    render_all(
        read_file("labels/{}.csv".format(dataset)),
        IMG_H,
        Path("images/{}/".format(dataset)),
    )



from random import sample, random, shuffle

values = ["jack", "queen", "king", "ace"]
suits = ["hearts", "clubs", "diamonds", "spades"]

deck = [(value, suit) for value in values for suit in suits]


def draw_cards(n_cards=4, n_decks=2):
    return list(sample(deck * n_decks, n_cards))


def weighted_sample(distribution):
    sample = random()
    for w, v in distribution:
        if sample < w:
            return v
        sample -= w
    return distribution[-1][1]


def next_game():
    cards = draw_cards()
    return cards[0:2], cards[2:4]


fair = [0.25, 0.25, 0.25, 0.25], [0.25, 0.25, 0.25, 0.25]

unfair = [0.2, 0.4, 0.15, 0.25], [0.25, 0.25, 0.25, 0.25]


def card_to_string(v, s):
    return "{} of {}".format(v, s)


def get_N(N, distribution):
    out = []
    p_tot = 1.0
    for p in distribution:
        n = int(round(p / p_tot * N))
        out.append(n)
        N -= n
        p_tot -= p
    # out.append(N - total)
    return out


def write_examples(distribution, N, location):
    lines = []
    house_cards = []
    for i, n1 in enumerate(get_N(N, distribution[0])):
        for j, n2 in enumerate(get_N(n1, distribution[1])):
            house_cards += [(values[i], suits[j])] * n2

    shuffle(house_cards)
    print(house_cards[:10])
    for house in house_cards:
        h1, h2 = next_game()
        h1, h2 = [card_to_string(*x) for x in h1], [card_to_string(*x) for x in h2]
        house = card_to_string(*house)
        lines.append(",".join(h1 + h2 + [house]))
        with open(location, "w") as f:
            f.write("\n".join(lines))


write_examples(fair, 500, "labels/fair.csv")
write_examples(unfair, 500, "labels/unfair.csv")

write_examples(fair, 100, "labels/fair_test.csv")
write_examples(unfair, 100, "labels/unfair_test.csv")



from typing import Dict, Tuple, Union

from problog.logic import (
    Term,
    Constant,
    list2term,
    term2list,
    Var,
    is_list,
    Clause,
    And,
    Or,
)
from problog.parser import PrologParser
from problog.program import ExtendedPrologFactory
from pyswip import Functor, Atom, Variable

PySwipObject = Union[Functor, Atom, Variable, int, float, list, bytes]
ProblogObject = Union[Term, Constant, Var, Clause, And, Or]


def pyswip_to_term(
    pyswip_obj: PySwipObject, with_variables=False
) -> Union[ProblogObject, Tuple[ProblogObject, Dict[Var, Variable]]]:
    variables = dict()
    if type(pyswip_obj) is Functor:
        args = []
        for a in pyswip_obj.args:
            args2, variables2 = pyswip_to_term(a, True)
            args.append(args2)
            variables.update(variables2)
        operator = pyswip_obj.name.get_value()
        if operator == ":-":
            new_term = Clause(*args)
        elif operator == ",":
            new_term = And.from_list(args)
        elif operator == ";":
            new_term = Or.from_list(args)
        else:
            new_term = Term(operator, *args)
    elif type(pyswip_obj) is Atom:
        new_term = Term(pyswip_obj.get_value())
    elif type(pyswip_obj) is int or type(pyswip_obj) is float:
        new_term = Constant(pyswip_obj)
    elif type(pyswip_obj) is list:
        lst = []
        for o in pyswip_obj:
            e, vars2 = pyswip_to_term(o, True)
            lst.append(e)
            variables.update(vars2)
        new_term = list2term(lst)
    elif type(pyswip_obj) is Variable:
        new_term = Var(
            pyswip_obj.chars if pyswip_obj.chars else f"Var{pyswip_obj.handle}"
        )
        variables = {new_term: pyswip_obj}
    elif type(pyswip_obj) is bytes:
        new_term = Constant(pyswip_obj.decode("utf-8"))
    else:
        raise Exception(f"Unhandled type {type(pyswip_obj)} from object {pyswip_obj}")
    if with_variables:
        return new_term, variables
    else:
        return new_term


def term_to_pyswip(term: ProblogObject) -> PySwipObject:
    if type(term) is Term:
        if is_list(term):
            return [term_to_pyswip(x) for x in term2list(term, False)]
        args = [term_to_pyswip(arg) for arg in term.args]
        if not args:
            return Atom(term.functor)

        functor = Functor(term.functor, arity=term.arity)
        return functor(*args)
    elif type(term) is Constant:
        return term.functor
    elif type(term) is Var:
        return Variable(name=term.name)
    else:
        raise Exception(
            f"Unhandled type {type(term)} from object {term} -> Robin has to fix it"
        )


_parser = PrologParser(ExtendedPrologFactory())


def parse(to_parse: Union[str, PySwipObject]) -> ProblogObject:
    if type(to_parse) is str:
        return _parser.parseString(str(to_parse) + ".")[0]
    return pyswip_to_term(to_parse)



from json import dumps

from deepproblog.engines import ApproximateEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.network import Network
from deepproblog.model import Model
from deepproblog.dataset import DataLoader
from deepproblog.examples.CLUTRR.architecture import Encoder, RelNet, GenderNet
from deepproblog.examples.CLUTRR.data import CLUTRR, CLUTRR_Dataset
from deepproblog.train import TrainObject
from deepproblog.utils.stop_condition import Threshold, StopOnPlateau
import torch

name = 'sys_gen_0'
clutrr = CLUTRR(name)

embed_size = 32
lstm = Encoder(clutrr.get_vocabulary(), embed_size, p_drop=0.0)
lstm_net = Network(lstm, "encoder", optimizer=torch.optim.Adam(lstm.parameters(), lr=1e-2))
rel_net = Network(RelNet(embed_size, 2 * embed_size), "rel_extract")
rel_net.optimizer = torch.optim.Adam(rel_net.parameters(), lr=1e-2)
gender_net = GenderNet(clutrr.get_vocabulary(), embed_size)
gender_net = Network(gender_net, "gender_net", optimizer=torch.optim.Adam(gender_net.parameters(), lr=1e-2))

model_filename = "model_forward.pl"
model = Model(model_filename, [rel_net, lstm_net, gender_net])
model.set_engine(ApproximateEngine(model, 1, ApproximateEngine.geometric_mean, exploration=True))

dataset: CLUTRR_Dataset = clutrr.get_dataset(".*train", gender=True, query_type="split")
val_dataset = dataset.subset(100)
test_datasets = clutrr.get_dataset(".*test", gender=True, query_type="split", separate=True)

loader = DataLoader(dataset, 4)

train_log = TrainObject(model)
train_log.train(
    loader,
    Threshold("Accuracy", 1.0) + StopOnPlateau("Accuracy", patience=5, warm_up=10),
    initial_test=False,
    test=lambda x: [("Accuracy", get_confusion_matrix(x, val_dataset, verbose=1).accuracy(),)],
    log_iter=50,
    test_iter=250,
)

model.save_state("models/" + name + ".pth")

for dataset in test_datasets:
    final_acc = get_confusion_matrix(
        model, test_datasets[dataset], verbose=0
    ).accuracy()
    train_log.logger.comment("{}\t{}".format(dataset, final_acc))
train_log.logger.comment(dumps(model.get_hyperparameters()))
train_log.write_to_file("log/" + name)



from time import time
from typing import Type, TYPE_CHECKING, Optional, List, Sequence

from problog.formula import LogicFormula
from problog.logic import Term

from deepproblog.arithmetic_circuit import ArithmeticCircuit
from deepproblog.engines import Engine
from deepproblog.query import Query
from deepproblog.semiring import Result
from deepproblog.semiring.graph_semiring import GraphSemiring, Semiring
from deepproblog.utils.cache import Cache

if TYPE_CHECKING:
    from deepproblog.model import Model


class SolverException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class Solver(object):
    """
    A class that bundles the different steps of inference.
    """

    def __init__(
        self,
        model: "Model",
        engine: Engine,
        semiring: Type[Semiring] = GraphSemiring,
        cache: bool = False,
        cache_root: Optional[str] = None,
        sdd_auto_gc: bool = False,
    ):
        """

        :param model: The model in which queries will be evaluated.
        :param engine: The engine that will be used to ground queries.
        :param semiring: The semiring that will be used to evaluate the arithmetic circuits.
        :param cache: If true, then arithmetic circuits will be cached.
        :param cache_root: If cache_root is not None, then the cache is persistent and is saved to that directory.
        :param sdd_auto_gc: Controls if the PySDD auto-GC feature should be turned on (may be needed for large problems)
        """
        if cache and not engine.can_cache():
            raise SolverException(
                "Caching is enabled, but {} cannot cache.".format(type(engine))
            )

        self.cache = Cache[Query, ArithmeticCircuit](
            func=self.build_ac,
            cache=cache,
            cache_root=cache_root,
            key_func=lambda x: x.query,
        )
        self.engine = engine
        self.model = model
        self.program = self.engine.prepare(model.program)
        self.semiring = semiring
        self.sdd_auto_gc = sdd_auto_gc

    def build_ac(self, q: Query) -> ArithmeticCircuit:
        """
        Builds the arithmetic circuit.
        :param q: The query for which to build the arithmetic circuit.
        :return: The arithmetic circuit for the given query.
        """
        start = time()

        ground = self.engine.ground(q, label=LogicFormula.LABEL_QUERY)
        ground_time = time() - start
        ac = ArithmeticCircuit(
            ground, self.semiring, ground_time=ground_time, sdd_auto_gc=self.sdd_auto_gc
        )
        return ac

    def solve(self, batch: Sequence[Query]) -> List[Result]:
        """
        Performs inference for a batch of queries.
        :param batch: A list of queries to perform inference on.
        :return: A list of results for the given queries.
        """
        self.engine.tensor_store.clear()
        # Build ACs
        acs: List[ArithmeticCircuit] = [self.cache.get(q) for q in batch]
        # Evaluate ACs. Evaluate networks if necessary
        result = [
            ac.evaluate(self.model, batch[i].substitution) for i, ac in enumerate(acs)
        ]
        semirings = [r.semiring for r in result]
        self.engine.perform_count(batch, (acs, semirings))
        return result

    def get_tensor(self, term: Term):
        return self.engine.get_tensor(term)

    def get_hyperparameters(self) -> dict:
        parameters = dict()
        parameters["engine"] = self.engine.get_hyperparameters()
        parameters["semiring"] = self.semiring.__name__
        return parameters



import torch

from data import MNISTImages, AdditionDataset
from deepproblog.dataset import DataLoader
from deepproblog.engines import ExactEngine
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model
from network import MNIST_Net

network = MNIST_Net()
net = Network(network, "mnist_net", batching=True)
net.optimizer = torch.optim.Adam(network.parameters(), lr=1e-3)

model = Model("addition.pl", [net])
model.set_engine(ExactEngine(model))
model.add_tensor_source("train", MNISTImages("train"))
model.add_tensor_source("test", MNISTImages("test"))

dataset = AdditionDataset("train")

# Train the model
loader = DataLoader(dataset, 2, False)
train_model(model, loader, 1, log_iter=100, profile=0)
model.save_state("snapshot/trained_model.pth")

# Query the model
query = dataset.to_query(0)
result = model.solve([query])[0]
print(result)



from deepproblog.dataset import DataLoader, QueryDataset
from deepproblog.engines import ExactEngine
from deepproblog.evaluate import get_confusion_matrix
from deepproblog.examples.Forth.WAP.wap_network import get_networks
from deepproblog.model import Model
from deepproblog.network import Network
from deepproblog.train import train_model

train_queries = QueryDataset("data/train.pl")
dev_queries = QueryDataset("data/dev.pl")
test_queries = QueryDataset("data/test.pl")

networks = get_networks(0.005, 0.5)

train_networks = [Network(x[0], x[1], x[2]) for x in networks]
test_networks = [Network(networks[0][0], networks[0][1])] + [
    Network(x[0], x[1], k=1) for x in networks[1:]
]

model = Model("wap.pl", train_networks)
model.set_engine(ExactEngine(model), cache=True)

test_model = Model("wap.pl", test_networks)
test_model.set_engine(ExactEngine(test_model), cache=False)

train_obj = train_model(
    model,
    DataLoader(train_queries, 10),
    40,
    log_iter=10,
    test=lambda x: [
        ("Accuracy", get_confusion_matrix(test_model, test_queries).accuracy())
    ],
    test_iter=30,
)



from typing import Optional, List, Sequence

from problog.logic import Term, Var


class Query(object):
    """
    The query object.
    """

    def __init__(
        self,
        query: Term,
        substitution: Optional[dict] = None,
        p: float = 1.0,
        output_ind: Sequence[int] = (-1,),
    ):
        """

        :param query: The query term.
        :param substitution: An optional substitution that can be applied to the term (see Term.apply_term).
        Used for parameterized queries.
        :param p: The target probability of this query.
        :param output_ind: A tuple that contains the indices of the arguments that should be considered output
        arguments. This is relevant for testing / negative mining.
        """
        self.query = query
        self.substitution = substitution
        if self.substitution is None:
            self.substitution = {}
        self.p = p
        self.output_ind = output_ind

    def variable_output(self) -> "Query":
        """

        :return:  A new query identical to the current one, but with the output arguments replaced with variables.
        """
        new_args = list(self.query.args)
        for i in self.output_ind:
            new_args[i] = Var("X_{}".format(i % len(new_args)))
        return Query(self.query(*new_args), self.substitution, self.p, self.output_ind)

    def replace_output(self, new_values: List[Term]) -> "Query":
        """
        Replaces the output variables
        :param new_values: The new values in order that should replace the output variables.
        :return: The query with the out_variables replaced by the corresponding new values.
        """
        new_args = list(self.query.args)
        j = 0
        for i in self.output_ind:
            new_args[i] = new_values[j]
            j += 1
        return Query(self.query(*new_args), self.substitution, self.p, self.output_ind)

    def output_values(self) -> List[Term]:
        """

        :return: The values of the output arguments
        """
        return [self.query.args[i] for i in self.output_ind]

    def substitute(self, substitution: Optional[dict] = None) -> "Query":
        """

        :param substitution: The dictionary that will be used to perform the substitution.
                             If None, then self.substitution will be used instead.
        :return: A new query where the substitution is applied. See the apply_term
                 method from the Term class for details on the substitution.
        """
        if substitution is None:
            substitution = self.substitution
        return Query(
            self.query.apply_term(substitution),
            substitution=None,
            p=self.p,
            output_ind=self.output_ind,
        )

    def __repr__(self):
        return "({}::{}, {})".format(self.p, self.query, self.substitution)

    def to_text_query(self):
        """Return the query on a form where QueryDataset can load it"""
        subst = sorted(
            ([k, v] for k, v in self.substitution.items()), key=lambda x: x[0].functor
        )
        return f"{self.p}::subs({self.query}, {subst})."

    def __hash__(self):
        return (
            hash(self.query)
            ^ hash(self.substitution)
            ^ hash(self.p)
            ^ hash(self.output_ind)
        )

    def __eq__(self, other):
        if not isinstance(other, Query):
            return False
        return (
            self.query == other.query
            and self.substitution == other.substitution
            and self.p == other.p
            and self.output_ind == other.output_ind
        )



from __future__ import annotations

from os import PathLike
from typing import Any, Dict, IO, Iterator, Optional, Union, List

import torch


def get_tensor_function(network: Network):
    def tensor_function(*args):
        return tuple(network.model.get_tensor(arg) for arg in args)

    return tensor_function


class Network(object):
    """Wraps a PyTorch neural network for use with DeepProblog"""

    def __init__(
            self,
            network_module: torch.nn.Module,
            name: str,
            optimizer: Optional[torch.optim.Optimizer] = None,
            scheduler=None,
            k: Optional[int] = None,
            batching: bool = False,
    ):
        """Create a Network object

        :param network_module: The neural network module.
        :param name: The name of the network as used in the neural predicate nn(name, ...)
        :param optimizer: The optimizer that updates the neural network parameters.
        :param scheduler: An optional learn rate scheduler for the optimizer.
        :param k: If k is set, only the top k results of the neural network will be used.
        :param batching: If batching is true, the inputs will be stacked and evaluated in the network in a batch.
                         Otherwise, they are evaluated one by one.
        """
        self.network_module = network_module
        self.name = name
        # self.function = function
        # if function is None:
        self.function = get_tensor_function(self)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.model = None
        self.is_cuda = False
        self.device = None
        self.n = 0
        self.domain = None
        self.cache = dict()
        self.k = k
        self.eval_mode = False
        self.batching = batching
        self.det = False

    def zero_grad(self):
        """
        Call zero grad on the optimizer
        :return:
        """
        if self.optimizer is not None:
            self.optimizer.zero_grad()

    def parameters(self) -> Iterator[torch.nn.Parameter]:
        """
        Return the parameters of the network module
        :return:
        """
        return self.network_module.parameters()

    def step(self):
        """
        Call the step function of the optimizer
        :return:
        """
        if self.optimizer is not None:
            self.optimizer.step()

    def step_epoch(self):
        """
        Call the step function of the scheduler
        :return:
        """
        if self.scheduler is not None:
            self.scheduler.step()

    def load(self, location: Union[PathLike, IO[bytes]]):
        """
        Restore the parameters of this network from a file.
        :param location: The location of the parameter file.
        :return:
        """
        loaded: Dict[str, Any] = torch.load(location)
        self.network_module.load_state_dict(loaded["model_state_dict"])
        if "optimizer_state_dict" in loaded:
            assert self.optimizer is not None
            self.optimizer.load_state_dict(loaded["optimizer_state_dict"])
        if "scheduler_state_dict" in loaded:
            assert self.scheduler is not None
            self.scheduler.load_state_dict(loaded["scheduler_state_dict"])

    def save(self, location: Union[PathLike, IO[bytes]], complete=False):
        """
        Save the parameters of this network to the given location.
        :param location: The location to save the parameter file to.
        :param complete: If true, also save information needed to resume training.
        :return:
        """
        to_save = {"model_state_dict": self.network_module.state_dict()}
        if complete:
            if self.optimizer is not None:
                to_save["optimizer_state_dict"] = self.optimizer.state_dict()
            if self.scheduler is not None:
                to_save["scheduler_state_dict"] = self.scheduler.state_dict()
        torch.save(to_save, location)

    def __call__(self, to_evaluate: list) -> list:
        """
        Evaluate the network on the given inputs.
        :param to_evaluate: A list that contains the inputs that the neural network should be evaluated on.
        :return:
        """
        if self.batching:
            inputs = (self.function(*e) for e in to_evaluate)
            stacked_inputs = list()
            for inputs in zip(*inputs):
                try:
                    inputs = torch.stack(inputs)
                    if self.is_cuda:
                        inputs.cuda(device=self.device)
                except TypeError:
                    inputs = list(inputs)
                stacked_inputs.append(inputs)
            evaluated = self.network_module(*stacked_inputs)
        else:
            evaluated = [self.network_module(*self.function(*e)) for e in to_evaluate]
        return evaluated

    def cuda(self, device=None):
        """
        Mode the network to a cuda device if CUDA is available.
        :param device:  The cuda device to move the network to.
        """
        if torch.cuda.is_available():
            self.is_cuda = True
            self.device = device
            self.network_module.cuda(device)
            print("Moving ", self.name, " to GPU")
        else:
            print("CUDA is unavailable")
        return self

    def eval(self):
        """
        Set the network to eval mode.
        """
        self.network_module.eval()
        self.eval_mode = True

    def train(self):
        """
        Set the network to train mode.
        """
        self.network_module.train()
        self.eval_mode = False

    def get_hyperparameters(self):
        parameters = {
            "name": self.name,
            "module": str(self.network_module),
            "optimizer": str(self.optimizer),
            "k": self.k,
        }
        return parameters

# class NetworkEvaluation(object):
#     """
#     An object that keeps track of which inputs the neural networks need to be evaluated on.
#     """
#
#     def __init__(self, batching: bool = False):
#         self.evaluated = None
#         self.added = []
#         self.batching = batching
#
#     def add(self, e):
#         self.added.append(e)
#
#     def evaluate(self, function: Callable, network: Network):
#         net = network.network_module
#         if self.batching:
#             batched_inputs = [function(*e) for e in self.added]
#             self.evaluated = net(torch.stack(batched_inputs))
#         else:
#             self.evaluated = torch.stack([net(*function(*e)) for e in self.added], 0)
#
#     def __getitem__(self, item):
#         return self.evaluated[item]
#
#     def item(self):
#         return self[0]
